import logging
import sys

import pytest
from lsprotocol.types import WorkspaceFolder
from pygls import IS_WIN
from pygls.workspace import Workspace

from esbonio.server.features.sphinx_manager.client import ClientState
from esbonio.server.features.sphinx_manager.client_subprocess import (
    SubprocessSphinxClient,
)
from esbonio.server.features.sphinx_manager.client_subprocess import (
    make_test_sphinx_client,
)
from esbonio.server.features.sphinx_manager.config import SphinxConfig
from esbonio.server.features.sphinx_manager.config import SubProcess

logger = logging.getLogger("__name__")


@pytest.mark.asyncio
async def test_create_application(uri_for):
    """Ensure that we can create a Sphinx application instance correctly."""

    demo_workspace = uri_for("workspaces", "demo")
    test_uri = demo_workspace / "index.rst"

    workspace = Workspace(
        None,
        workspace_folders=[
            WorkspaceFolder(uri=str(demo_workspace), name="demo"),
        ],
    )
    config = SphinxConfig(
        python_command=SubProcess(command=[sys.executable]),
        config_overrides={
            "html_theme": "alabaster",
            "html_theme_options": {},
        },
    )
    resolved = config.resolve(test_uri, workspace, logger)
    assert resolved is not None
    client = None

    try:
        client = await make_test_sphinx_client(resolved)
        assert client.state == ClientState.Running

        info = client.sphinx_info
        assert info is not None
        assert info.builder_name == "dirhtml"

        # Paths are case insensitive on Windows
        if IS_WIN:
            assert info.src_dir.lower() == demo_workspace.fs_path.lower()
            assert info.conf_dir.lower() == demo_workspace.fs_path.lower()
            assert "cache" in info.build_dir.lower()
        else:
            assert info.src_dir == demo_workspace.fs_path
            assert info.conf_dir == demo_workspace.fs_path
            assert "cache" in info.build_dir
    finally:
        if client:
            await client.stop()


@pytest.mark.asyncio
async def test_create_application_error(uri_for, tmp_path_factory):
    """Ensure that we can handle errors during application creation."""

    build_dir = tmp_path_factory.mktemp("build")
    demo_workspace = uri_for("workspaces", "demo")
    test_uri = demo_workspace / "index.rst"

    workspace = Workspace(
        None,
        workspace_folders=[
            WorkspaceFolder(uri=str(demo_workspace), name="demo"),
        ],
    )

    conf_dir = uri_for("workspaces", "demo-error").fs_path
    config = SphinxConfig(
        python_command=SubProcess(command=[sys.executable]),
        build_command=[
            "sphinx-build",
            "-b",
            "html",
            "-c",
            conf_dir,
            demo_workspace.fs_path,
            str(build_dir),
        ],
        config_overrides={
            "html_theme": "alabaster",
            "html_theme_options": {},
        },
    )
    resolved = config.resolve(test_uri, workspace, logger)
    assert resolved is not None

    try:
        client = await SubprocessSphinxClient(resolved)
        assert client.state == ClientState.Errored

        message = "There is a programmable error in your configuration file:"
        assert message in str(client.exception)
    finally:
        await client.stop()



from __future__ import annotations

import os
import pathlib
import typing

from lsprotocol import types as lsp

from esbonio import server

if typing.TYPE_CHECKING:
    from collections.abc import Coroutine
    from typing import Any


class RoleTargetProvider:
    """Base class for role target providers."""

    def __init__(self, esbonio: server.EsbonioLanguageServer):
        self.converter = esbonio.converter
        self.logger = esbonio.logger.getChild(self.__class__.__name__)

    def hover_target(
        self, context: server.HoverContext, target: str, **kwargs
    ) -> str | None | Coroutine[Any, Any, str | None]:
        """Return the hover text for the given target"""
        return None

    def suggest_targets(
        self, context: server.CompletionContext, **kwargs
    ) -> (
        list[lsp.CompletionItem]
        | None
        | Coroutine[Any, Any, list[lsp.CompletionItem] | None]
    ):
        """Givem a completion context, suggest role targets that may be used."""
        return None

    def find_target_definition(
        self, context: server.DefinitionContext, target: str, **kwargs
    ) -> list[lsp.Location] | None:
        """Find the definition(s) for the given role target.

        Parameters
        ----------
        context
           The context of the definition request

        target
           The target to find the definition for
        """
        return None

    def resolve_target_link(
        self, context: server.DocumentLinkContext, target: str, **kwargs
    ) -> (
        str
        | None
        | tuple[str, str | None]
        | Coroutine[Any, Any, str | None | tuple[str, str | None]]
    ):
        """Resolve a document link request for the given role target.

        Parameters
        ----------
        context
           The context of the document link request

        target
           The target to resolve the link for
        """
        return None


class FilepathProvider(RoleTargetProvider):
    """Target provider for filepaths."""

    def resolve_target_link(
        self,
        context: server.DocumentLinkContext,
        target: str,
        *,
        root: str = "/",
        pattern: str | None = None,
        **kwargs,
    ) -> None | str | tuple[str, str | None]:
        """Given a directive's argument, resolve a link for it if possible.

        Parameters
        ----------
        root
           If the user provides an absolute path, resolve the link relative to this directory.
        """
        path = self._resolve_path(context, target, root)
        uri = server.Uri.for_file(os.path.normpath(path))
        tooltip = "Path exists" if path.exists() else "Path does NOT exist"

        return str(uri), tooltip

    def find_target_definition(
        self,
        context: server.DefinitionContext,
        target: str,
        *,
        root: str = "/",
        pattern: str | None = None,
        **kwargs,
    ) -> list[lsp.Location] | None:
        path = self._resolve_path(context, target, root)
        uri = server.Uri.for_file(os.path.normpath(path))

        if not uri:
            return None

        return [
            lsp.Location(
                uri=str(uri),
                range=lsp.Range(
                    start=lsp.Position(line=0, character=0),
                    end=lsp.Position(line=1, character=0),
                ),
            )
        ]

    def suggest_targets(
        self,
        context: server.CompletionContext,
        *,
        root: str = "/",
        pattern: str | None = None,
        **kwargs,
    ) -> list[lsp.CompletionItem]:
        """Given a completion context, suggest files (or folders) that may be used.

        Parameters
        ----------
        root
           If the user provides an absolute path, generate suggestions relative to this directory.

        pattern
           If set, limit suggestions only to matching files.

        Returns
        -------
        list[lsp.CompletionItem]
           A list of completion items to suggest.
        """
        uri = context.uri
        cwd = pathlib.Path(uri).parent

        if (partial := context.match.group("label")) and partial.startswith("/"):
            candidate_dir = pathlib.Path(root)

            # Be sure to remove the leading '/', otherwise partial will wipe out the
            # root when concatenated.
            partial = partial[1:]
        else:
            candidate_dir = cwd

        candidate_dir /= partial
        if partial and not partial.endswith(("/", ".")):
            candidate_dir = candidate_dir.parent

        self.logger.debug("Suggesting files relative to %r", candidate_dir)
        return [
            self._path_to_completion_item(context, p) for p in candidate_dir.glob("*")
        ]

    def _path_to_completion_item(
        self, context: server.CompletionContext, path: pathlib.Path
    ) -> lsp.CompletionItem:
        """Create the ``CompletionItem`` for the given path.

        In the case where there are multiple filepath components, this function needs to
        provide an appropriate ``TextEdit`` so that the most recent entry in the path can
        be easily edited - without clobbering the existing path.

        Also bear in mind that this function must play nice with both role target and
        directive argument completions.
        """
        new_text = f"{path.name}"
        kind = (
            lsp.CompletionItemKind.Folder
            if path.is_dir()
            else lsp.CompletionItemKind.File
        )

        # TODO: Mkae this work
        # if (start := self._find_start_char(context)) == -1:
        #     insert_text = new_text
        #     filter_text = None
        #     text_edit = None
        # else:
        #     start += 1
        #     _, end = context.match.span()
        #     prefix = context.match.group(0)[start:end]

        #     self.logger.debug(f"{context.match.group(0)}, {start=}, {end=}, {prefix=}")

        #     insert_text = None
        #     filter_text = f"{prefix}{new_text}"  # Needed so VSCode will actually show the results.

        #     text_edit = lsp.TextEdit(
        #         range=lsp.Range(
        #             start=lsp.Position(line=context.position.line, character=start),
        #             end=lsp.Position(line=context.position.line, character=end),
        #         ),
        #         new_text=new_text,
        #     )

        return lsp.CompletionItem(
            label=new_text,
            kind=kind,
            # insert_text=insert_text,
            # filter_text=filter_text,
            # text_edit=text_edit,
        )

    def _find_start_char(self, context: server.CompletionContext) -> int:
        matched_text = context.match.group(0)
        idx = matched_text.find("/")

        while True:
            next_idx = matched_text.find("/", idx + 1)
            if next_idx == -1:
                break

            idx = next_idx

        return idx

    def _resolve_path(
        self, context: server.UriContext, argument: str, root: str
    ) -> pathlib.Path:
        if argument.startswith("/"):
            # Be sure to remove the leading '/', otherwise `argument` will wipe out the
            # root when concatenated.
            path = pathlib.Path(root, argument[1:])
        else:
            cwd = pathlib.Path(context.uri).parent
            path = cwd / argument

        return path



from __future__ import annotations

import functools
import logging
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from types import TracebackType
from typing import Any
from typing import TypeVar

from . import types
from .util import send_message

T = TypeVar("T")


def patch_sphinx():
    """Monkey patch parts of Sphinx with our own implementations."""
    try:
        from sphinx.util import display
    except ImportError:
        # display submodule was not introduced until Sphinx 6.x
        import sphinx.util as display  # type: ignore[no-redef]

    display.status_iterator = status_iterator
    display.progress_message = progress_message  # type: ignore


def status_iterator(
    iterable: Iterable[T],
    summary: str,
    color: str = "",
    length: int = 0,
    verbosity: int = 0,
    stringify_func: Callable[[Any], str] | None = None,
) -> Iterator[T]:
    """Used to override Sphinx's version of this function.
    Sends progress reports to the client as well as the usual logs.
    """
    from sphinx.util.logging import NAMESPACE

    try:
        from sphinx.util.display import display_chunk
    except ImportError:
        # display submodule was not introduced until Sphinx 5.x
        from sphinx.util import display_chunk  # type: ignore[no-redef]

    if stringify_func is None:
        stringify_func = display_chunk

    logger = logging.getLogger(NAMESPACE)
    send_message(types.ProgressMessage(params=types.ProgressParams(message=summary)))

    if verbosity == 0:
        logger.info(summary)

    percentage = " "
    for i, item in enumerate(iterable, start=1):
        if verbosity > 0:
            if length > 0:
                percentage = f" [{int((i / length) * 100): >3d}%] "
            logger.info(f"{summary}{percentage}{stringify_func(item)}")

        yield item


class progress_message:
    """Used to override Spinx's version of this, reports progress to the client."""

    def __init__(self, message: str, *, nonl: bool = True) -> None:
        self.message = message

    def __enter__(self) -> None:
        send_message(
            types.ProgressMessage(params=types.ProgressParams(message=self.message))
        )

    def __exit__(
        self,
        typ: type[BaseException] | None,
        val: BaseException | None,
        tb: TracebackType | None,
    ) -> bool:
        from sphinx.locale import __
        from sphinx.util.logging import NAMESPACE

        try:
            from sphinx.util import display
        except ImportError:
            # display submodule was not introduced until Sphinx 6.x
            import sphinx.util as display  # type: ignore[no-redef]

        logger = logging.getLogger(NAMESPACE)

        if isinstance(val, display.SkipProgressMessage):
            logger.info(self.message + "... " + __("skipped"))
            if val.args:
                logger.info(*val.args)
            return True
        elif val:
            logger.info(__(self.message + "... " + "failed"))
        else:
            logger.info(self.message + "... " + __("done"))

        return False

    def __call__(self, f: Callable) -> Callable:
        @functools.wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with self:
                return f(*args, **kwargs)

        return wrapper



import asyncio
import pathlib
import sys

import pytest
import pytest_lsp
from lsprotocol import types
from pytest_lsp import ClientServerConfig
from pytest_lsp import LanguageClient
from sphinx import version_info as sphinx_version

SERVER_CMD = ["-m", "esbonio"]
TEST_DIR = pathlib.Path(__file__).parent.parent


@pytest.mark.asyncio(loop_scope="session")
async def test_rst_document_diagnostic(client: LanguageClient, uri_for):
    """Ensure that we can get the diagnostics for a single rst document correctly."""

    workspace_uri = uri_for("workspaces", "demo")
    test_uri = workspace_uri / "rst" / "diagnostics.rst"
    report = await client.text_document_diagnostic_async(
        types.DocumentDiagnosticParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri))
        )
    )

    assert report.kind == "full"

    category = " [ref.ref]" if sphinx_version[0] >= 8 else ""
    message = f"undefined label: 'not-a-real-reference'{category}"

    # We will only check the diagnostic message, full details will be handled by other
    # test cases.
    messages = {d.message for d in report.items}
    assert messages == {message}

    assert len(client.diagnostics) == 0, "Server should not publish diagnostics"


@pytest.mark.asyncio(loop_scope="session")
async def test_myst_document_diagnostic(client: LanguageClient, uri_for):
    """Ensure that we can get the diagnostics for a single myst document correctly."""

    workspace_uri = uri_for("workspaces", "demo")
    test_uri = workspace_uri / "myst" / "diagnostics.md"
    report = await client.text_document_diagnostic_async(
        types.DocumentDiagnosticParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri))
        )
    )

    assert report.kind == "full"

    category = " [ref.ref]" if sphinx_version[0] >= 8 else ""
    message = f"undefined label: 'not-a-real-reference'{category}"

    # We will only check the diagnostic message, full details will be handled by other
    # test cases.
    messages = {d.message for d in report.items}
    assert messages == {message}

    assert len(client.diagnostics) == 0, "Server should not publish diagnostics"


@pytest.mark.asyncio(loop_scope="session")
async def test_workspace_diagnostic(client: LanguageClient, uri_for):
    """Ensure that we can get diagnostics for the whole workspace correctly."""
    report = await client.workspace_diagnostic_async(
        types.WorkspaceDiagnosticParams(previous_result_ids=[])
    )

    category = " [ref.ref]" if sphinx_version[0] >= 8 else ""
    message = f"undefined label: 'not-a-real-reference'{category}"

    workspace_uri = uri_for("workspaces", "demo")
    expected = {
        str(workspace_uri / "conf.py"): (
            "no theme named 'furo' found",
            "Could not import extension sphinx_design (exception: "
            "No module named 'sphinx_design')",
        ),
        str(workspace_uri / "index.rst"): ('Unknown directive type "grid"'),
        str(workspace_uri / "rst" / "diagnostics.rst"): (message,),
        str(workspace_uri / "myst" / "diagnostics.md"): (message,),
    }
    assert len(report.items) == len(expected)
    for item in report.items:
        for diagnostic in item.items:
            assert diagnostic.message.startswith(expected[item.uri])

    assert len(client.diagnostics) == 0, "Server should not publish diagnostics"


@pytest_lsp.fixture(
    scope="module",
    config=ClientServerConfig(
        server_command=[sys.executable, *SERVER_CMD],
    ),
)
async def pub_client(lsp_client: LanguageClient, uri_for, tmp_path_factory):
    """A client that does **not** support the pull-diagnostics model."""

    workspace_uri = uri_for("workspaces", "demo")
    test_uri = workspace_uri / "rst" / "diagnostics.rst"

    # Override the project configuration so that builds are written into a temporary folder
    build_dir = tmp_path_factory.mktemp("build")
    lsp_client.set_configuration({"logging": {"level": "debug"}}, section="esbonio")
    lsp_client.set_configuration(
        {
            "buildCommand": [
                "sphinx-build",
                "-M",
                "html",
                workspace_uri.fs_path,
                str(build_dir),
            ],
            "configOverrides": {
                "html_theme": "alabaster",
                "html_theme_options": {},
            },
            "pythonCommand": [sys.executable],
        },
        section="esbonio.sphinx",
        scope_uri=str(workspace_uri),
    )

    await lsp_client.initialize_session(
        types.InitializeParams(
            capabilities=types.ClientCapabilities(
                # Signal workDoneProgress/create support.
                window=types.WindowClientCapabilities(
                    work_done_progress=True,
                ),
                workspace=types.WorkspaceClientCapabilities(configuration=True),
            ),
            workspace_folders=[
                types.WorkspaceFolder(uri=str(workspace_uri), name="demo"),
            ],
        )
    )

    # Open a text document to trigger sphinx client creation.
    lsp_client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=str(test_uri),
                language_id="restructuredtext",
                version=0,
                text=pathlib.Path(test_uri).read_text(),
            )
        )
    )

    await lsp_client.wait_for_notification("sphinx/appCreated")

    # Save the document to trigger a build
    lsp_client.text_document_did_save(
        types.DidSaveTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri))
        )
    )

    build_finished = False
    while not build_finished:
        await lsp_client.wait_for_notification("$/progress")
        report = list(lsp_client.progress_reports.values())[0][-1]
        build_finished = report.kind == "end"

    yield

    # Teardown
    await asyncio.wait_for(lsp_client.shutdown_session(), timeout=2.0)


@pytest.mark.asyncio(loop_scope="module")
async def test_publish_diagnostics(pub_client: LanguageClient, uri_for):
    """Ensure that the server publishes the diagnostics it finds"""
    workspace_uri = uri_for("workspaces", "demo")

    category = " [ref.ref]" if sphinx_version[0] >= 8 else ""
    message = f"undefined label: 'not-a-real-reference'{category}"

    expected = {
        str(workspace_uri / "rst" / "diagnostics.rst"): {message},
        str(workspace_uri / "myst" / "diagnostics.md"): {message},
    }

    # The server might not have published its diagnostics yet
    while len(pub_client.diagnostics) < len(expected):
        await pub_client.wait_for_notification(types.TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)

    for uri, expected_msgs in expected.items():
        items = pub_client.diagnostics[uri]
        actual_msgs = {d.message for d in items}

        assert expected_msgs == actual_msgs



"""Type definitions for the sphinx agent.

This is the *only* module shared between the agent itself and the parent language
server. For this reason this module *cannot* import anything from Sphinx.
"""

from __future__ import annotations

import dataclasses
from typing import Any

from .directives import MYST_DIRECTIVE
from .directives import RST_DIRECTIVE
from .directives import RST_DIRECTIVE_OPTION
from .directives import Directive
from .lsp import Diagnostic
from .lsp import DiagnosticSeverity
from .lsp import Location
from .lsp import Position
from .lsp import Range
from .roles import MYST_ROLE
from .roles import RST_DEFAULT_ROLE
from .roles import RST_ROLE
from .roles import Role
from .uri import IS_WIN
from .uri import Uri

__all__ = (
    "Diagnostic",
    "DiagnosticSeverity",
    "Directive",
    "IS_WIN",
    "Location",
    "MYST_DIRECTIVE",
    "MYST_ROLE",
    "Position",
    "RST_DEFAULT_ROLE",
    "RST_DIRECTIVE",
    "RST_DIRECTIVE_OPTION",
    "RST_ROLE",
    "Range",
    "Role",
    "Uri",
)


# -- DB Types
#
# These represent the structure of data as stored in the SQLite database
Symbol = tuple[  # Represents either a document symbol or workspace symbol depending on context.
    int,  # id
    str,  # name
    int,  # kind
    str,  # detail
    str,  # range - as json object
    int | None,  # parent_id
    int,  # order_id
]


# -- RPC Types
#
# These represent the structure of the messages sent between the Sphinx agent and the
# parent language server.
@dataclasses.dataclass
class CreateApplicationParams:
    """Parameters of a ``sphinx/createApp`` request."""

    command: list[str]
    """The ``sphinx-build`` command to base the app instance on."""

    config_overrides: dict[str, Any]
    """Overrides to apply to the application's configuration."""

    context: dict[str, str] = dataclasses.field(default_factory=dict)
    """The context in which to resolve config variables."""


@dataclasses.dataclass
class CreateApplicationRequest:
    """A ``sphinx/createApp`` request."""

    id: int | str

    params: CreateApplicationParams

    method: str = "sphinx/createApp"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class SphinxInfo:
    """Represents information about an instance of the Sphinx application."""

    version: str
    """The version of Sphinx being used."""

    conf_dir: str
    """The folder containing the project's conf.py"""

    build_dir: str
    """The folder containing the Sphinx application's build output"""

    builder_name: str
    """The name of the builder in use"""

    src_dir: str
    """The folder containing the source files for the project"""

    dbpath: str
    """The filepath the database is stored in."""


@dataclasses.dataclass
class CreateApplicationResponse:
    """A ``sphinx/createApp`` response."""

    id: int | str

    result: SphinxInfo

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class BuildParams:
    """Parameters of a ``sphinx/build`` request."""

    filenames: list[str] = dataclasses.field(default_factory=list)

    force_all: bool = False

    content_overrides: dict[str, str] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class BuildResult:
    """Results from a ``sphinx/build`` request."""

    diagnostics: dict[str, list[Diagnostic]] = dataclasses.field(default_factory=dict)
    """Any diagnostics associated with the project."""


@dataclasses.dataclass
class BuildRequest:
    """A ``sphinx/build`` request."""

    id: int | str

    params: BuildParams

    method: str = "sphinx/build"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class BuildResponse:
    """A ``sphinx/build`` response."""

    id: int | str

    result: BuildResult

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class LogMessageParams:
    """Parameters of a ``window/logMessage`` notification."""

    type: int

    message: str


@dataclasses.dataclass
class LogMessage:
    """A ``window/logMessage`` notification"""

    params: LogMessageParams

    method: str = "window/logMessage"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class ProgressParams:
    message: str | None = None

    percentage: int | None = None


@dataclasses.dataclass
class ProgressMessage:
    """A ``$/progress`` notification"""

    params: ProgressParams

    method: str = "$/progress"

    jsonrpc: str = dataclasses.field(default="2.0")


@dataclasses.dataclass
class ExitNotification:
    """An ``exit`` notification"""

    params: None

    method: str = "exit"

    jsonrpc: str = dataclasses.field(default="2.0")


METHOD_TO_MESSAGE_TYPE = {
    BuildRequest.method: BuildRequest,
    ExitNotification.method: ExitNotification,
    CreateApplicationRequest.method: CreateApplicationRequest,
}
METHOD_TO_RESPONSE_TYPE = {
    BuildRequest.method: BuildResponse,
    ExitNotification.method: None,
    CreateApplicationRequest.method: CreateApplicationResponse,
}



from __future__ import annotations

import ast

import pytest
from lsprotocol import types

from esbonio.server.testing import range_from_str
from esbonio.sphinx_agent.app import find_extension_declaration
from esbonio.sphinx_agent.app import find_html_theme_declaration

CONF_PY = """\
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from docutils import nodes
from sphinx.application import Sphinx

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Esbonio Demo"
copyright = "2023, Esbonio Developers"
author = "Esbonio Developers"
release = "1.0"

extensions = [ "a",
  "sphinx.ext.intersphinx",
      "myst-parser",
]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_title = "Esbonio Demo"
html_theme_options = {
    "source_repository": "https://github.com/swyddfa/esbonio",
    "source_branch": "develop",
    "source_directory": "lib/esbonio/tests/workspaces/demo/",
}
"""


@pytest.mark.parametrize(
    "src,expected",
    [
        ("", None),
        ("a=3", None),
        (CONF_PY, range_from_str("23:0-23:19")),
    ],
)
def test_find_html_theme_declaration(src: str, expected: types.Range | None):
    """Ensure that we can locate the location within a ``conf.py``
    file where the ``html_theme`` is defined."""

    mod = ast.parse(src)
    actual = find_html_theme_declaration(mod)

    if expected is None:
        assert actual is None

    else:
        assert actual.start.line == expected.start.line
        assert actual.end.line == expected.end.line

        assert actual.start.character == expected.start.character
        assert actual.end.character == expected.end.character


@pytest.mark.parametrize(
    "src,extname,expected",
    [
        ("", "myst-parser", None),
        ("a=3", "myst-parser", None),
        ("extensions='a'", "myst-parser", None),
        ("extensions=['myst-parser']", "myst-parser", range_from_str("0:12-0:25")),
        (CONF_PY, "myst-parser", range_from_str("17:6-17:19")),
    ],
)
def test_find_extension_declaration(
    src: str, extname: str, expected: types.Range | None
):
    """Ensure that we can locate the location within a ``conf.py``
    file where the ``html_theme`` is defined."""

    mod = ast.parse(src)
    actual = find_extension_declaration(mod, extname)

    if expected is None:
        assert actual is None

    else:
        assert actual.start.line == expected.start.line
        assert actual.end.line == expected.end.line

        assert actual.start.character == expected.start.character
        assert actual.end.character == expected.end.character



from __future__ import annotations

import json
import pathlib
import typing

import aiosqlite

from esbonio.server import Uri
from esbonio.sphinx_agent import types

if typing.TYPE_CHECKING:
    from typing import Any
    from typing import TypeVar

    import cattrs

    T = TypeVar("T")


class Project:
    """Represents a documentation project."""

    def __init__(self, dbpath: str | pathlib.Path, converter: cattrs.Converter):
        self.converter = converter
        self.dbpath = dbpath
        self._connection: aiosqlite.Connection | None = None

    async def close(self):
        if self._connection is not None:
            await self._connection.close()

    async def get_db(self) -> aiosqlite.Connection:
        if self._connection is None:
            self._connection = await aiosqlite.connect(self.dbpath)

        return self._connection

    def load_as(self, o: str, t: type[T]) -> T:
        return self.converter.structure(json.loads(o), t)

    async def uri_to_docname(self, uri: str | Uri) -> str | None:
        """Given a uri, look up the corresponding Sphinx docname."""
        db = await self.get_db()

        query = "SELECT docname FROM files WHERE uri = ?"
        cursor = await db.execute(query, (str(uri),))

        if (result := await cursor.fetchone()) is None:
            return None

        return result[0]

    async def docname_to_uri(self, docname: str) -> str | None:
        """Given a Sphinx docname, lookup the corresponding uri"""
        db = await self.get_db()

        query = "SELECT uri FROM files WHERE docname = ?"
        cursor = await db.execute(query, (docname,))

        if (result := await cursor.fetchone()) is None:
            return None

        return result[0]

    async def get_src_uris(self) -> list[Uri]:
        """Return all known source uris."""
        db = await self.get_db()

        query = "SELECT uri FROM files"
        async with db.execute(query) as cursor:
            results = await cursor.fetchall()
            return [Uri.parse(s[0]) for s in results]

    async def get_build_path(self, src_uri: Uri) -> str | None:
        """Get the build path associated with the given ``src_uri``."""
        db = await self.get_db()

        query = "SELECT urlpath FROM files WHERE uri = ?"
        async with db.execute(query, (str(src_uri.resolve()),)) as cursor:
            if (result := await cursor.fetchone()) is None:
                return None

            return result[0]

    async def get_config_value(self, name: str) -> Any | None:
        """Return the requested configuration value, if available."""

        db = await self.get_db()
        query = "SELECT value FROM config WHERE name = ?"
        cursor = await db.execute(query, (name,))

        if (row := await cursor.fetchone()) is None:
            return None

        (value,) = row
        return json.loads(value)

    async def get_directive(self, name: str) -> types.Directive | None:
        """Get the directive with the given name."""
        db = await self.get_db()

        query = "SELECT * FROM directives WHERE name = ?"
        cursor = await db.execute(query, (name,))
        result = await cursor.fetchone()

        return (
            types.Directive.from_db(self.load_as, *result)
            if result is not None
            else None
        )

    async def get_directives(self) -> list[tuple[str, str | None]]:
        """Get the directives known to Sphinx."""
        db = await self.get_db()

        query = "SELECT name, implementation FROM directives"
        cursor = await db.execute(query)
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_role(self, name: str) -> types.Role | None:
        """Get the role with the given name."""
        db = await self.get_db()

        query = "SELECT * FROM roles WHERE name = ?"
        cursor = await db.execute(query, (name,))
        result = await cursor.fetchone()

        return types.Role.from_db(self.load_as, *result) if result is not None else None

    async def get_roles(self) -> list[tuple[str, str | None]]:
        """Get the roles known to Sphinx."""
        db = await self.get_db()

        query = "SELECT name, implementation FROM roles"
        cursor = await db.execute(query)
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_document_symbols(self, src_uri: Uri) -> list[types.Symbol]:
        """Get the symbols for the given file."""
        db = await self.get_db()
        query = (
            "SELECT id, name, kind, detail, range, parent_id, order_id "
            "FROM symbols WHERE uri = ?"
        )
        cursor = await db.execute(query, (str(src_uri.resolve()),))
        return await cursor.fetchall()  # type: ignore[return-value]

    async def find_symbols(self, **kwargs) -> list[types.Symbol]:
        """Find symbols which match the given criteria."""
        db = await self.get_db()
        base_query = (
            "SELECT id, name, kind, detail, range, parent_id, order_id FROM symbols"
        )
        where: list[str] = []
        parameters: list[Any] = []

        for param, value in kwargs.items():
            where.append(f"{param} = ?")
            parameters.append(value)

        if where:
            conditions = " AND ".join(where)
            query = " ".join([base_query, "WHERE", conditions])
        else:
            query = base_query

        cursor = await db.execute(query, tuple(parameters))
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_workspace_symbols(
        self, query: str
    ) -> list[tuple[str, str, int, str, str, str]]:
        """Return all the workspace symbols matching the given query string"""

        db = await self.get_db()
        sql_query = """\
SELECT
    child.uri,
    child.name,
    child.kind,
    child.detail,
    child.range,
    COALESCE(parent.name, '') AS container_name
FROM
    symbols child
LEFT JOIN
    symbols parent ON (child.parent_id = parent.id AND child.uri = parent.uri)
WHERE
    child.name like ? or child.detail like ?;"""

        query_str = f"%{query}%"
        cursor = await db.execute(sql_query, (query_str, query_str))
        return await cursor.fetchall()  # type: ignore[return-value]

    async def get_diagnostics(self) -> dict[Uri, list[dict[str, Any]]]:
        """Get diagnostics for the project."""
        db = await self.get_db()
        cursor = await db.execute("SELECT * FROM diagnostics")
        results: dict[Uri, list[dict[str, Any]]] = {}

        for uri_str, item in await cursor.fetchall():
            uri = Uri.parse(uri_str)
            diagnostic = json.loads(item)
            results.setdefault(uri, []).append(diagnostic)

        return results



from __future__ import annotations

import typing

from docutils import nodes
from sphinx import addnodes

from .. import types
from ..app import Database
from ..app import Sphinx
from ..util import as_json

if typing.TYPE_CHECKING:
    from sphinx.domains import Domain
    from sphinx.util.typing import Inventory


PROJECTS_TABLE = Database.Table(
    "intersphinx_projects",
    [
        Database.Column(name="id", dtype="TEXT"),
        Database.Column(name="name", dtype="TEXT"),
        Database.Column(name="version", dtype="TEXT"),
        Database.Column(name="uri", dtype="TEXT"),
    ],
)


OBJECTS_TABLE = Database.Table(
    "objects",
    [
        Database.Column(name="name", dtype="TEXT"),
        Database.Column(name="display", dtype="TEXT"),
        Database.Column(name="domain", dtype="TEXT"),
        Database.Column(name="objtype", dtype="TEXT"),
        Database.Column(name="docname", dtype="TEXT"),
        Database.Column(name="project", dtype="TEXT"),
        Database.Column(name="description", dtype="TEXT"),
        Database.Column(name="location", dtype="JSON"),
    ],
)


class DomainObjects:
    """Discovers and indexes domain objects."""

    def __init__(self, app: Sphinx):
        self._info: dict[tuple[str, str, str, str], tuple[str | None, str | None]] = {}

        # Needs to run late, but before the handler in ./roles.py
        app.connect("builder-inited", self.init_db, priority=998)

        app.connect("doctree-read", self.doctree_read)
        app.connect("object-description-transform", self.object_defined)
        app.connect("build-finished", self.commit)

    def init_db(self, app: Sphinx):
        """Prepare the database."""
        projects = index_intersphinx_projects(app)
        project_names = [p[0] for p in projects]

        for domain in app.env.domains.values():
            index_domain_directives(app, domain)
            index_domain_roles(app, domain, project_names)

    def doctree_read(self, app: Sphinx, doctree: nodes.document):
        """Extract information from the given doctree.

        Currently only used to get the location of ``:ref:`` targets
        """
        for node in doctree.traverse(condition=nodes.target):
            if (label := node.attributes.get("refid")) is None:
                continue

            if (source := node.source) is None or (line := node.line) is None:
                continue

            location = as_json(
                types.Location(
                    uri=str(types.Uri.for_file(source)),
                    range=types.Range(
                        start=types.Position(line=line - 1, character=0),
                        end=types.Position(line=line, character=0),
                    ),
                )
            )

            key = (label, "std", "label", app.env.docname)
            self._info[key] = ("", location)

    def commit(self, app, exc):
        """Commit changes to the database.

        The only way to guarantee we discover all objects, from all domains correctly,
        is to call the ``get_objects()`` method on each domain. This means we process
        every object, every time we build.

        I will be *very* surprised if this never becomes a performance issue, but we
        will have to think of a smarter approach when it comes to it.
        """
        app.esbonio.db.clear_table(OBJECTS_TABLE, project=None)
        rows = []

        for name, domain in app.env.domains.items():
            for objname, dispname, objtype, docname, _, _ in domain.get_objects():
                desc, location = self._get_object_details(
                    app, objname, name, objtype, docname
                )

                if objname == (display := str(dispname)):
                    display = "-"

                rows.append(
                    (objname, display, name, objtype, docname, None, desc, location)
                )

        app.esbonio.db.insert_values(OBJECTS_TABLE, rows)
        # self._info.clear()

    def _get_object_details(
        self, app: Sphinx, objname: str, domain: str, objtype: str, docname: str
    ) -> tuple[str | None, str | None]:
        """Get additional details about the given object."""
        desc, location = self._info.get(
            (objname, domain, objtype, docname), (None, None)
        )

        if location is None and f"{domain}:{objtype}" == "std:doc":
            docpath = app.env.doc2path(objname)
            location = as_json(
                types.Location(
                    uri=str(types.Uri.for_file(docpath)),
                    range=types.Range(
                        start=types.Position(line=0, character=0),
                        end=types.Position(line=1, character=0),
                    ),
                )
            )

        return desc, location

    def object_defined(
        self, app: Sphinx, domain: str, objtype: str, content: addnodes.desc_content
    ):
        """Record additional information about a domain object.

        Despite having a certain amount of structure to them (thanks to the API),
        domains can still do arbitrary things - take a peek at the implementations of
        the ``std``, ``py`` and ``cpp`` domains!

        So while this will never be perfect, this method is called each time the
        ``object-description-transform`` event is fired and attempts to extract the
        object's description and precise location.

        The trick however, is linking these items up with the correct object
        """

        sig = content.parent[0]

        try:
            name = sig["ids"][0]  # type: ignore[index]
        except Exception:
            return

        docname = app.env.docname
        description = content.astext()

        if (source := sig.source) is not None and (line := sig.line) is not None:
            location = as_json(
                types.Location(
                    uri=str(types.Uri.for_file(source)),
                    range=types.Range(
                        start=types.Position(line=line - 1, character=0),
                        end=types.Position(line=line, character=0),
                    ),
                )
            )
        else:
            location = None

        key = (name, domain, objtype, docname)
        self._info[key] = (description, location)


def index_domain_directives(app: Sphinx, domain: Domain):
    """Index the directives in the given domain.

    Parameters
    ----------
    app
       The application instance

    domain
       The domain to index
    """
    for name, directive in domain.directives.items():
        app.esbonio.add_directive(f"{domain.name}:{name}", directive, [])


def index_domain_roles(app: Sphinx, domain: Domain, projects: list[str] | None):
    """Index the roles in the given domain.

    Parameters
    ----------
    app
       The application instance

    domain
       The domain to index

    projects
       The list of known intersphinx projects
    """
    target_types: dict[str, set[str]] = {}

    for obj_name, item_type in domain.object_types.items():
        for role_name in item_type.roles:
            target_type = f"{domain.name}:{obj_name}"
            target_types.setdefault(role_name, set()).add(target_type)

    for name, role in domain.roles.items():
        if (item_types := target_types.get(name)) is None:
            app.esbonio.add_role(f"{domain.name}:{name}", role, [])
            continue

        # Add an entry for the local project.
        provider = app.esbonio.create_role_target_provider(
            "objects", obj_types=list(item_types), projects=None
        )
        app.esbonio.add_role(f"{domain.name}:{name}", role, [provider])

        if projects is None or len(projects) == 0:
            continue

        # Add an entry referencing all external projects
        provider = app.esbonio.create_role_target_provider(
            "objects", obj_types=list(item_types), projects=projects
        )
        app.esbonio.add_role(f"external:{domain.name}:{name}", role, [provider])

        # Add an entry dedicated to each external project
        for project in projects:
            provider = app.esbonio.create_role_target_provider(
                "objects", obj_types=list(item_types), projects=[project]
            )
            app.esbonio.add_role(
                f"external+{project}:{domain.name}:{name}", role, [provider]
            )


def index_intersphinx_projects(app: Sphinx) -> list[tuple[str, str, str, str]]:
    """Index all the projects known to intersphinx.

    Parameters
    ----------
    app
       The application instance

    Returns
    -------
    List[Tuple[str, str, str, str]]
       The list of discovered projects
    """
    app.esbonio.db.ensure_table(OBJECTS_TABLE)
    app.esbonio.db.ensure_table(PROJECTS_TABLE)
    app.esbonio.db.clear_table(PROJECTS_TABLE)

    projects: list[tuple[str, str, str, str]] = []
    objects = []

    mapping = getattr(app.config, "intersphinx_mapping", {})
    inventory = getattr(app.env, "intersphinx_named_inventory", {})

    for id_, (_, (uri, _)) in mapping.items():
        if (project := inventory.get(id_, None)) is None:
            continue

        app.esbonio.db.clear_table(OBJECTS_TABLE, project=id_)

        # We just need an entry to be able to extract the project name and version
        (name, version, _, _) = next(iter(next(iter(project.values())).values()))

        projects.append((id_, name, version, uri))
        objects.extend(index_intersphinx_objects(id_, uri, project))

    app.esbonio.db.insert_values(PROJECTS_TABLE, projects)
    app.esbonio.db.insert_values(OBJECTS_TABLE, objects)

    return projects


def index_intersphinx_objects(project_name: str, uri: str, project: Inventory):
    """Index all the objects in the given project."""

    objects = []

    for objtype, items in project.items():
        domain = None
        if ":" in objtype:
            domain, *parts = objtype.split(":")
            objtype = ":".join(parts)

        for objname, (_, _, item_uri, display) in items.items():
            docname = item_uri.replace(uri, "")
            objects.append(
                (objname, display, domain, objtype, docname, project_name, None, None)
            )

    return objects


def setup(app: Sphinx):
    DomainObjects(app)



from __future__ import annotations

import importlib
import inspect

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives as docutils_directives
from pygments.lexers import get_all_lexers

from .. import types
from ..app import Database
from ..app import Sphinx
from ..util import as_json

DIRECTIVES_TABLE = Database.Table(
    "directives",
    [
        Database.Column(name="name", dtype="TEXT"),
        Database.Column(name="implementation", dtype="TEXT"),
        Database.Column(name="location", dtype="JSON"),
        Database.Column(name="argument_providers", dtype="JSON"),
    ],
)


def get_impl_name(directive: type[Directive]) -> str:
    try:
        return f"{directive.__module__}.{directive.__name__}"
    except AttributeError:
        return f"{directive.__module__}.{directive.__class__.__name__}"


def get_impl_location(impl: type[Directive]) -> str | None:
    """Get the implementation location of the given directive"""

    try:
        if (filepath := inspect.getsourcefile(impl)) is None:
            return None

        uri = types.Uri.for_file(filepath).resolve()
        source, line = inspect.getsourcelines(impl)

        location = types.Location(
            uri=str(uri),
            range=types.Range(
                start=types.Position(line=line - 1, character=0),
                end=types.Position(line=line + len(source), character=0),
            ),
        )

        return as_json(location)
    except Exception:
        # TODO: Log the error somewhere..
        return None


def index_directives(app: Sphinx):
    """Index all the directives that are available to this app.

    Note: While it would be ideal to resolve the implementation location of each
    directive here, the ``get_impl_location`` function is too slow causing a noticable
    lag when initializing the sphinx agent.

    Perhaps it's worth investigating adding our own custom events. If we could register
    a handler and do work on an "initial" doctree when it becomes available, then we can
    resolve the location of just the directives that are used, as we see them for the
    first time.
    """

    directives: dict[str, types.Directive] = {}

    # Process the roles registered through Sphinx
    for name, impl, providers in app.esbonio._directives:
        directives[name] = types.Directive(
            name, get_impl_name(impl), argument_providers=providers
        )

    ignored_directives = {"restructuredtext-test-directive"}
    found_directives = {
        **docutils_directives._directive_registry,  # type: ignore[attr-defined]
        **docutils_directives._directives,  # type: ignore[attr-defined]
    }

    for name, directive in found_directives.items():
        if name in ignored_directives or name in directives:
            continue

        # core docutils directives are a (module, Class) reference.
        if isinstance(directive, tuple):
            try:
                mod, cls = directive
                modulename = f"docutils.parsers.rst.directives.{mod}"
                module = importlib.import_module(modulename)

                directive = getattr(module, cls)
            except Exception:
                # TODO: Log the error somewhere...
                directives[name] = types.Directive(name, None, None)
                continue

        directives[name] = types.Directive(name, get_impl_name(directive))

    _add_providers(app, directives)

    app.esbonio.db.ensure_table(DIRECTIVES_TABLE)
    app.esbonio.db.clear_table(DIRECTIVES_TABLE)
    app.esbonio.db.insert_values(
        DIRECTIVES_TABLE, [d.to_db(as_json) for d in directives.values()]
    )


def setup(app: Sphinx):
    app.connect("builder-inited", index_directives, priority=999)


def _add_providers(app: Sphinx, directives: dict[str, types.Directive]):
    """Add provider definitions to built in directive types we know about."""

    lexers_provider = _get_lexers_provider()
    filepath_provider = types.Directive.ArgumentProvider(
        "filepath", {"root": app.srcdir}
    )

    for name in ["code-block", "sourcecode", "highlight"]:
        if (directive := directives.get(name)) is not None:
            directive.argument_providers = [lexers_provider]

    for name in ["image", "figure", "include", "literalinclude"]:
        if (directive := directives.get(name)) is not None:
            directive.argument_providers = [filepath_provider]


def _get_lexers_provider() -> types.Directive.ArgumentProvider:
    """Get the argument provider instance that returns the names of pygments lexers."""

    langs = []
    for name, labels, files, mimes in get_all_lexers():
        filenames = ", ".join(f"`{f}`" for f in files)
        mimetypes = ", ".join(f"`{m}`" for m in mimes)

        for label in labels:
            langs.append(
                {
                    "label": label,
                    "kind": 21,  # Constant
                    "documentation": {
                        "kind": "markdown",
                        "value": f"### {name}\nFilenames: {filenames}\n\nMIME Types: {mimetypes}",
                    },
                }
            )

    return types.Directive.ArgumentProvider("values", {"values": langs})



from __future__ import annotations

import typing

from esbonio.sphinx_agent.types import Uri

from ._configuration import ConfigChangeEvent
from ._configuration import ConfigurationContext
from .events import EventSource
from .feature import CompletionConfig
from .feature import CompletionContext
from .feature import CompletionTrigger
from .feature import DefinitionContext
from .feature import DefinitionTrigger
from .feature import DocumentLinkContext
from .feature import HoverContext
from .feature import HoverTrigger
from .feature import LanguageFeature
from .server import EsbonioLanguageServer
from .server import EsbonioWorkspace
from .server import __version__
from .setup import create_language_server

if typing.TYPE_CHECKING:
    from .feature import UriContext  # noqa: F401

__all__ = (
    "__version__",
    "ConfigChangeEvent",
    "ConfigurationContext",
    "CompletionConfig",
    "CompletionContext",
    "CompletionTrigger",
    "DefinitionContext",
    "DefinitionTrigger",
    "DocumentLinkContext",
    "EsbonioLanguageServer",
    "EsbonioWorkspace",
    "EventSource",
    "HoverContext",
    "HoverTrigger",
    "LanguageFeature",
    "Uri",
    "create_language_server",
)



from __future__ import annotations

from lsprotocol import types

from esbonio import server
from esbonio.server.features.roles import RolesFeature
from esbonio.server.features.roles import completion
from esbonio.sphinx_agent.types import MYST_ROLE


class MystRoles(server.LanguageFeature):
    """A frontend to roles for MyST syntax."""

    def __init__(self, roles: RolesFeature, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.roles = roles
        self._insert_behavior = "replace"

    completion_trigger = server.CompletionTrigger(
        patterns=[MYST_ROLE],
        languages={"markdown"},
        characters={"{", "`", "<", "/"},
    )

    definition_trigger = server.DefinitionTrigger(
        patterns=[MYST_ROLE],
        languages={"markdown"},
    )

    def initialized(self, params: types.InitializedParams):
        """Called once the initial handshake between client and server has finished."""
        self.configuration.subscribe(
            "esbonio.server.completion",
            server.CompletionConfig,
            self.update_configuration,
        )

    def update_configuration(
        self, event: server.ConfigChangeEvent[server.CompletionConfig]
    ):
        """Called when the user's configuration is updated."""
        self._insert_behavior = event.value.preferred_insert_behavior

    async def completion(
        self, context: server.CompletionContext
    ) -> list[types.CompletionItem] | None:
        """Provide completion suggestions for roles."""

        groups = context.match.groupdict()
        target = groups["target"]

        # All text matched by the regex
        text = context.match.group(0)
        start, end = context.match.span()

        if target:
            target_index = start + text.find(target)

            # Only trigger target completions if the request was made from within
            # the target part of the role.
            if target_index <= context.position.character <= end:
                return await self.complete_targets(context)

        return await self.complete_roles(context)

    async def complete_targets(self, context: server.CompletionContext):
        """Provide completion suggestions for role targets."""

        render_func = completion.get_role_target_renderer(
            context.language, self._insert_behavior
        )
        if render_func is None:
            return None

        items = []
        role_name = context.match.group("name")
        for target in await self.roles.suggest_targets(context, role_name):
            if (item := render_func(context, target)) is not None:
                items.append(item)

        return items if len(items) > 0 else None

    async def complete_roles(
        self, context: server.CompletionContext
    ) -> list[types.CompletionItem] | None:
        """Return completion suggestions for the available roles"""

        render_func = completion.get_role_renderer(
            context.language, self._insert_behavior
        )
        if render_func is None:
            return None

        items = []
        for role in await self.roles.suggest_roles(context):
            if (item := render_func(context, role)) is not None:
                items.append(item)

        if len(items) > 0:
            return items

        return None

    async def definition(
        self, context: server.DefinitionContext
    ) -> list[types.Location] | None:
        """Find the definition of the requested item"""
        role = context.match.group("name")
        target = context.match.group("target")
        label = context.match.group("label")

        if not label:
            return None

        idx = context.match.group(0).index(target)
        start = context.match.start() + idx
        end = start + len(target)

        if start <= context.position.character <= end:
            return await self.roles.find_target_definition(context, role, label)

        return None

    async def document_link(
        self, context: server.DocumentLinkContext
    ) -> list[types.DocumentLink] | None:
        links = []

        for line, text in enumerate(context.doc.lines):
            for match in MYST_ROLE.finditer(text):
                if not (target := match.group("label")):
                    continue

                name = match.group("name")
                link_target = await self.roles.resolve_target_link(
                    context, name, target
                )

                if link_target is None:
                    continue

                tooltip = None
                if isinstance(link_target, tuple):
                    link_target, tooltip = link_target

                char = "<" if match.group("alias") is not None else "`"
                search = f"{char}{target}"

                idx = match.group(0).index(search) + 1
                start = match.start() + idx
                end = start + len(target)

                links.append(
                    types.DocumentLink(
                        target=link_target,
                        tooltip=tooltip if context.tooltip_support else None,
                        range=types.Range(
                            start=types.Position(line=line, character=start),
                            end=types.Position(line=line, character=end),
                        ),
                    )
                )

        return links if len(links) > 0 else None

    hover_trigger = server.HoverTrigger(
        patterns=[MYST_ROLE],
        languages={"markdown"},
    )

    async def hover(self, context: server.HoverContext) -> types.Hover | None:
        """Find the hover text of the requested item"""
        role = context.match.group("name")
        target = context.match.group("target")
        label = context.match.group("label")

        if not label:
            return None

        idx = context.match.group(0).index(target)
        start = context.match.start() + idx
        end = start + len(target)

        if start <= context.position.character <= end:
            if (text := await self.roles.hover_target(context, role, label)) is None:
                return None
        else:
            return None

        linum = context.position.line
        return types.Hover(
            contents=types.MarkupContent(kind=types.MarkupKind.Markdown, value=text),
            range=types.Range(
                start=types.Position(line=linum, character=start),
                end=types.Position(line=linum, character=end),
            ),
        )


def esbonio_setup(esbonio: server.EsbonioLanguageServer, roles: RolesFeature):
    rst_roles = MystRoles(roles, esbonio)
    esbonio.add_feature(rst_roles)



from __future__ import annotations

from esbonio.server import EsbonioLanguageServer

from .manager import ProjectManager
from .project import Project

__all__ = [
    "Project",
    "ProjectManager",
]


def esbonio_setup(server: EsbonioLanguageServer):
    manager = ProjectManager(server)
    server.add_feature(manager)



from __future__ import annotations

import typing
from typing import Literal

import attrs
from lsprotocol import types
from pygls.capabilities import get_capability
from pygls.workspace import TextDocument

from . import Uri

if typing.TYPE_CHECKING:
    import re
    from collections.abc import Coroutine
    from typing import Any
    from typing import Protocol

    from .server import EsbonioLanguageServer

    CompletionResult = (
        None
        | list[types.CompletionItem]
        | Coroutine[Any, Any, list[types.CompletionItem] | None]
    )

    DefinitionResult = (
        None | list[types.Location] | Coroutine[Any, Any, list[types.Location] | None]
    )

    DocumentLinkResult = (
        None
        | list[types.DocumentLink]
        | Coroutine[Any, Any, list[types.DocumentLink] | None]
    )

    DocumentSymbolResult = (
        None
        | list[types.DocumentSymbol]
        | Coroutine[Any, Any, list[types.DocumentSymbol] | None]
    )

    HoverResult = None | types.Hover | Coroutine[Any, Any, types.Hover | None]
    MaybeAsyncNone = None | Coroutine[Any, Any, None]

    WorkspaceSymbolResult = (
        None
        | list[types.WorkspaceSymbol]
        | Coroutine[Any, Any, list[types.WorkspaceSymbol] | None]
    )

    class UriContext(Protocol):
        uri: Uri


class LanguageFeature:
    """Base class for language features."""

    def __init__(self, server: EsbonioLanguageServer):
        self.server = server
        self.logger = server.logger.getChild(self.__class__.__name__)

    @property
    def converter(self):
        return self.server.converter

    @property
    def configuration(self):
        return self.server.configuration

    def initialize(self, params: types.InitializeParams) -> MaybeAsyncNone:
        """Called during ``initialize``."""

    def initialized(self, params: types.InitializedParams) -> MaybeAsyncNone:
        """Called when the ``initialized`` notification is received."""

    def shutdown(self, params: None) -> MaybeAsyncNone:
        """Called when the server is instructed to ``shutdown`` by the client."""

    def document_change(
        self, params: types.DidChangeTextDocumentParams
    ) -> MaybeAsyncNone:
        """Called when a text document is changed."""

    def document_close(
        self, params: types.DidCloseTextDocumentParams
    ) -> MaybeAsyncNone:
        """Called when a text document is closed."""

    def document_open(self, params: types.DidOpenTextDocumentParams) -> MaybeAsyncNone:
        """Called when a text document is opened."""

    def document_save(self, params: types.DidSaveTextDocumentParams) -> MaybeAsyncNone:
        """Called when a text document is saved."""

    completion_trigger: CompletionTrigger | None = None

    def completion(self, context: CompletionContext) -> CompletionResult:
        """Called when a completion request matches one of the specified triggers."""

    definition_trigger: DefinitionTrigger | None = None

    def definition(self, context: DefinitionContext) -> DefinitionResult:
        """Called when a definition request matches one of the specified triggers."""

    hover_trigger: HoverTrigger | None = None

    def hover(self, context: HoverContext) -> HoverResult:
        """Called when a hover request matches one of the specified triggers."""

    def document_link(self, context: DocumentLinkContext) -> DocumentLinkResult:
        """Called when a document link request is recieved."""

    def document_symbol(
        self, params: types.DocumentSymbolParams
    ) -> DocumentSymbolResult:
        """Called when a document symbols request is received."""

    def workspace_symbol(
        self, params: types.WorkspaceSymbolParams
    ) -> WorkspaceSymbolResult:
        """Called when a workspace symbols request is received."""


@attrs.define
class CompletionTrigger:
    """Define when the feature's completion method should be called."""

    patterns: list[re.Pattern]
    """A list of regular expressions to try"""

    languages: set[str] = attrs.field(factory=set)
    """Languages in which the completion trigger should fire.

    If empty, the document's language will be ignored.
    """

    characters: set[str] = attrs.field(factory=set)
    """Characters which, when typed, should trigger a completion request.

    If empty, this trigger will ignore any trigger characters.
    """

    def __call__(
        self,
        uri: Uri,
        params: types.CompletionParams,
        document: TextDocument,
        language: str,
        client_capabilities: types.ClientCapabilities,
    ) -> CompletionContext | None:
        """Determine if this completion trigger should fire.

        Parameters
        ----------
        uri
           The uri of the document in which the completion request was made

        params
           The completion params sent from the client

        document
           The document in which the completion request was made

        language
           The language at the point where the completion request was made

        client_capabilities
           The client's capabilities

        Returns
        -------
        Optional[CompletionContext]
           A completion context, if this trigger has fired
        """

        if len(self.languages) > 0 and language not in self.languages:
            return None

        if not self._trigger_characters_match(params):
            return None

        try:
            line = document.lines[params.position.line]
        except IndexError:
            line = ""

        for pattern in self.patterns:
            for match in pattern.finditer(line):
                # Only trigger completions if the position of the request is within the
                # match.
                start, stop = match.span()
                if not (start <= params.position.character <= stop):
                    continue

                return CompletionContext(
                    uri=uri,
                    doc=document,
                    match=match,
                    position=params.position,
                    language=language,
                    capabilities=client_capabilities,
                )

        return None

    def _trigger_characters_match(self, params: types.CompletionParams) -> bool:
        """Determine if this trigger's completion characters align with the request."""

        if (context := params.context) is None:
            # No context available, assume a match
            return True

        if context.trigger_kind != types.CompletionTriggerKind.TriggerCharacter:
            # Not a trigger character request, assume a match
            return True

        if (char := context.trigger_character) is None or len(self.characters) == 0:
            return True

        return char in self.characters


@attrs.define
class CompletionConfig:
    """Configuration options that control completion behavior."""

    preferred_insert_behavior: Literal["insert", "replace"] = attrs.field(
        default="replace"
    )
    """This option indicates if the user prefers we use ``insertText`` or ``textEdit``
    when rendering ``CompletionItems``."""


@attrs.define
class CompletionContext:
    """Captures the context within which a completion request has been made."""

    uri: Uri
    """The uri for the document in which the completion request was made."""

    doc: TextDocument
    """The document within which the completion request was made."""

    match: re.Match
    """The match object describing the site of the completion request."""

    position: types.Position
    """The position at which the completion request was made."""

    language: str
    """The language where the completion request was made."""

    capabilities: types.ClientCapabilities
    """The client's capabilities."""

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return f"CompletionContext<{self.uri}:{p} -- {self.match}>"

    @property
    def commit_characters_support(self) -> bool:
        """Indicates if the client supports commit characters."""
        return get_capability(
            self.capabilities,
            "text_document.completion.completion_item.commit_characters_support",
            False,
        )

    @property
    def deprecated_support(self) -> bool:
        """Indicates if the client supports the deprecated field on a
        ``CompletionItem``."""
        return get_capability(
            self.capabilities,
            "text_document.completion.completion_item.deprecated_support",
            False,
        )

    @property
    def documentation_formats(self) -> list[types.MarkupKind]:
        """The list of documentation formats supported by the client."""
        return list(
            get_capability(
                self.capabilities,
                "text_document.completion.completion_item.documentation_format",
                [],
            )
        )

    @property
    def insert_replace_support(self) -> bool:
        """Indicates if the client supports ``InsertReplaceEdit``."""
        return get_capability(
            self.capabilities,
            "text_document.completion.completion_item.insert_replace_support",
            False,
        )

    @property
    def preselect_support(self) -> bool:
        """Indicates if the client supports the preselect field on a
        ``CompletionItem``."""
        return get_capability(
            self.capabilities,
            "text_document.completion.completion_item.preselect_support",
            False,
        )

    @property
    def snippet_support(self) -> bool:
        """Indicates if the client supports snippets"""
        return get_capability(
            self.capabilities,
            "text_document.completion.completion_item.snippet_support",
            False,
        )

    @property
    def supported_tags(self) -> list[types.CompletionItemTag]:
        """The list of ``CompletionItemTags`` supported by the client."""
        capabilities = get_capability(
            self.capabilities,
            "text_document.completion.completion_item.tag_support",
            None,
        )

        if not capabilities:
            return []

        return list(capabilities.value_set)


@attrs.define
class DefinitionTrigger:
    """Define when the feature's definition method should be called."""

    patterns: list[re.Pattern]
    """A list of regular expressions to try"""

    languages: set[str] = attrs.field(factory=set)
    """Languages in which the completion trigger should fire.

    If empty, the document's language will be ignored.
    """

    def __call__(
        self,
        uri: Uri,
        params: types.DefinitionParams,
        document: TextDocument,
        language: str,
        client_capabilities: types.ClientCapabilities,
    ) -> DefinitionContext | None:
        """Determine if this definition trigger should fire.

        Parameters
        ----------
        uri
           The uri of the document in which the request was made

        params
           The definition params sent from the client

        document
           The document in which the request was made

        language
           The language at the point where the request was made

        client_capabilities
           The client's capabilities

        Returns
        -------
        Optional[DefinitionContext]
           A definition context, if this trigger has fired
        """

        if len(self.languages) > 0 and language not in self.languages:
            return None

        try:
            line = document.lines[params.position.line]
        except IndexError:
            line = ""

        for pattern in self.patterns:
            for match in pattern.finditer(line):
                # Only trigger if the position of the request is within the match.
                start, stop = match.span()
                if not (start <= params.position.character <= stop):
                    continue

                return DefinitionContext(
                    uri=uri,
                    doc=document,
                    match=match,
                    position=params.position,
                    language=language,
                    capabilities=client_capabilities,
                )

        return None


@attrs.define
class DefinitionContext:
    """Captures the context within which a definition request has been made."""

    uri: Uri
    """The uri for the document in which the definition request was made"""

    doc: TextDocument
    """The document within which the definition request was made"""

    match: re.Match
    """The match object describing the site of the definition request."""

    position: types.Position
    """The position at which the definition request was made."""

    language: str
    """The language where the definition request was made."""

    capabilities: types.ClientCapabilities
    """The client's capabilities."""

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return f"DefinitionContext<{self.uri}:{p} ({self.language}) -- {self.match}>"


@attrs.define
class DocumentLinkContext:
    """Captures the context within which a document link request has been made."""

    uri: Uri
    """The uri for the document in which the completion request was made."""

    doc: TextDocument
    """The document within which the document link request was made."""

    capabilities: types.ClientCapabilities
    """The client's capabilities"""

    def __repr__(self):
        return f"DocumentLinkContext<{self.uri}>"

    @property
    def tooltip_support(self) -> bool:
        """Indicates if the client supports tooltips."""
        return get_capability(
            self.capabilities,
            "text_document.document_link.tooltip_support",
            False,
        )


@attrs.define
class HoverTrigger:
    """Define when the feature's hover method should be called."""

    patterns: list[re.Pattern]
    """A list of regular expressions to try"""

    languages: set[str] = attrs.field(factory=set)
    """Languages in which the trigger should fire.

    If empty, the document's language will be ignored.
    """

    def __call__(
        self,
        uri: Uri,
        params: types.HoverParams,
        document: TextDocument,
        language: str,
        client_capabilities: types.ClientCapabilities,
    ) -> HoverContext | None:
        """Determine if this hover trigger should fire.

        Parameters
        ----------
        uri
           The uri of the document in which the request was made

        params
           The definition params sent from the client

        document
           The document in which the request was made

        language
           The language at the point where the request was made

        client_capabilities
           The client's capabilities

        Returns
        -------
        Optional[HoverContext]
           A hover context, if this trigger has fired
        """

        if len(self.languages) > 0 and language not in self.languages:
            return None

        try:
            line = document.lines[params.position.line]
        except IndexError:
            line = ""

        for pattern in self.patterns:
            for match in pattern.finditer(line):
                # Only trigger if the position of the request is within the match.
                start, stop = match.span()
                if not (start <= params.position.character <= stop):
                    continue

                return HoverContext(
                    uri=uri,
                    doc=document,
                    match=match,
                    position=params.position,
                    language=language,
                    capabilities=client_capabilities,
                )

        return None


@attrs.define
class HoverContext:
    """Captures the context within which a hover request has been made."""

    uri: Uri
    """The uri for the document in which the hover request was made"""

    doc: TextDocument
    """The document within which the hover request was made"""

    match: re.Match
    """The match object describing the site of the hover request."""

    position: types.Position
    """The position at which the hover request was made."""

    language: str
    """The language where the hover request was made."""

    capabilities: types.ClientCapabilities
    """The client's capabilities."""

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return f"HoverContext<{self.uri}:{p} ({self.language}) -- {self.match}>"

    @property
    def content_format(self) -> list[types.MarkupKind]:
        """The list of supported markup formats the client supports (if known).
        Order indicates the client's preference"""

        hover: types.HoverClientCapabilities | None
        hover = get_capability(self.capabilities, "text_document.hover", None)
        if hover is None:
            return []

        return list(hover.content_format or [])

    @property
    def markdown_parser(self) -> tuple[str, str | None] | None:
        """The markdown parser used by the client (if known)"""

        markdown: types.MarkdownClientCapabilities | None
        markdown = get_capability(self.capabilities, "general.markdown", None)
        if markdown is None:
            return None

        return (markdown.parser, markdown.version)

    @property
    def markdown_allowed_tags(self) -> list[str]:
        """The list of allowed html tags the client will allow in markdown text (if
        known)"""

        markdown: types.MarkdownClientCapabilities | None
        markdown = get_capability(self.capabilities, "general.markdown", None)
        if markdown is None:
            return []

        return list(markdown.allowed_tags or [])



from typing import Any
from urllib.parse import urlencode

import attrs
from lsprotocol import types
from pygls.capabilities import get_capability

from esbonio import server
from esbonio.server import Uri
from esbonio.server.features.project_manager import ProjectManager
from esbonio.server.features.sphinx_manager import SphinxClient
from esbonio.server.features.sphinx_manager import SphinxManager

from .config import PreviewConfig
from .preview import PreviewServer
from .preview import make_http_server
from .webview import WebviewServer
from .webview import make_ws_server


@attrs.define
class PreviewFileParams:
    uri: str


class PreviewManager(server.LanguageFeature):
    """Language feature for managing previews."""

    def __init__(
        self,
        server: server.EsbonioLanguageServer,
        sphinx: SphinxManager,
        projects: ProjectManager,
    ):
        super().__init__(server)
        self.sphinx = sphinx
        self.sphinx.add_listener("build", self.on_build)
        """The sphinx manager."""

        self.built_clients: set[str] = set()
        """Keeps track of which clients run a build at least once."""

        self.build_path: str | None = None
        """The filepath we are currently displaying."""

        self.build_uri: Uri | None = None
        """The uri of the build dir we are currently serving from."""

        self.config = PreviewConfig()
        """The current configuration."""

        self.projects = projects
        """The project manager."""

        self.preview: PreviewServer | None = None
        """The http server for serving the built files"""

        self.webview: WebviewServer | None = None
        """The server for controlling the webview."""

    @property
    def supports_show_document(self):
        """Indicates if the client supports the `window/showDocument` request."""
        return get_capability(
            self.server.client_capabilities, "window.show_document.support", False
        )

    def initialized(self, params: types.InitializedParams):
        """Called once the initial handshake between client and server has finished."""
        self.configuration.subscribe(
            "esbonio.preview", PreviewConfig, self.update_configuration
        )

    def shutdown(self, params: None):
        """Called when the client instructs the server to ``shutdown``."""

        if self.preview is not None:
            self.preview.stop()

        if self.webview is not None:
            self.webview.stop()

    @property
    def preview_active(self) -> bool:
        """Return true if the preview is active.

        i.e. there is a HTTP server hosting the build result."""
        return self.preview is not None

    @property
    def preview_controllable(self) -> bool:
        """Return true if the preview is controllable.

        i.e. there is a web socket server available to control the webview.
        """
        return self.webview is not None

    def update_configuration(self, event: server.ConfigChangeEvent[PreviewConfig]):
        """Called when the user's configuration is updated."""
        config = event.value

        # (Re)create the websocket server
        if self.webview is None:
            self.webview = make_ws_server(self.server, config)

        elif (
            config.bind != self.webview.config.bind
            or config.ws_port != self.webview.config.ws_port
        ):
            self.webview.stop()
            self.webview = make_ws_server(self.server, config)

        # (Re)create the http server
        if self.preview is None:
            self.preview = make_http_server(self.server, config)
            self.preview.build_uri = self.build_uri

        elif (
            config.bind != self.preview.config.bind
            or config.http_port != self.preview.config.http_port
        ):
            self.preview.stop()
            self.preview = make_http_server(self.server, config)
            self.preview.build_uri = self.build_uri

        self.config = config
        self.webview.config = config
        self.server.run_task(self.show_preview_uri())

    async def on_build(self, client: SphinxClient, result):
        """Called whenever a sphinx build completes."""
        self.built_clients.add(client.id)

        if self.webview is None or self.preview is None:
            return

        # Only refresh the view if the project we are previewing was built.
        if client.build_uri != self.preview.build_uri:
            return

        self.logger.debug("Refreshing preview")
        self.webview.reload()

    async def scroll_view(self, uri: str, line: int):
        """Scroll the webview to the given line number."""

        if self.webview is None:
            return

        self.webview.scroll(uri, line)

    async def preview_file(self, params: PreviewFileParams, retry=True):
        if self.preview is None:
            return None

        # Always check the fully resolved uri.
        src_uri = Uri.parse(params.uri).resolve()
        self.logger.debug("Previewing file: '%s'", src_uri)

        if (client := await self.sphinx.get_client(src_uri)) is None:
            return None

        if (project := self.projects.get_project(src_uri)) is None:
            return None

        if (build_path := await project.get_build_path(src_uri)) is None:
            # The client might not have built the project yet.
            if client.id not in self.built_clients and retry is True:
                # Only retry this once.
                await self.sphinx.trigger_build(src_uri)
                return await self.preview_file(params, retry=False)
            else:
                self.logger.debug(
                    "Unable to preview file '%s', not included in build output.",
                    src_uri,
                )
                return None

        self.build_path = build_path
        self.build_uri = self.preview.build_uri = client.build_uri

        if (uri := await self.show_preview_uri()) is None:
            return None

        return {"uri": uri.as_string(encode=False)}

    async def show_preview_uri(self) -> Uri | None:
        """Show the preview uri in the client using a ``window/showDocument`` request.
        Also return the final uri."""

        if self.webview is None or self.preview is None or self.build_path is None:
            return None

        server = await self.preview
        webview = await self.webview

        host = "localhost"
        query_params: dict[str, Any] = dict(ws=f"ws://{host}:{webview.port}")

        if self.config.show_line_markers:
            query_params["show-markers"] = True

        uri = Uri.create(
            scheme="http",
            authority=f"{host}:{server.port}",
            path=self.build_path,
            query=urlencode(query_params),
        )
        self.logger.info("Preview available at: %s", uri.as_string(encode=False))

        if self.supports_show_document:
            result = await self.server.window_show_document_async(
                types.ShowDocumentParams(
                    uri=uri.as_string(encode=False), external=True, take_focus=False
                )
            )
            self.logger.debug("window/showDocument: %s", result)

        return uri


def esbonio_setup(
    esbonio: server.EsbonioLanguageServer,
    sphinx: SphinxManager,
    projects: ProjectManager,
):
    manager = PreviewManager(esbonio, sphinx, projects)
    esbonio.add_feature(manager)

    @esbonio.feature("view/scroll")
    async def on_scroll(ls: server.EsbonioLanguageServer, params):
        await manager.scroll_view(params.uri, params.line)

    @esbonio.command("esbonio.server.previewFile")
    async def preview_file(ls: server.EsbonioLanguageServer, params: PreviewFileParams):
        return await manager.preview_file(params)



from __future__ import annotations

import dataclasses
import os
import pathlib
import re
from collections.abc import Callable
from urllib import parse

IS_WIN = os.name == "nt"
SCHEME = re.compile(r"^[a-zA-Z][a-zA-Z\d+.-]*$")
RE_DRIVE_LETTER_PATH = re.compile(r"^(\/?)([a-zA-Z]:)")


# TODO: Look into upstreaming this into pygls
#       - if it works out
#       - when pygls drops 3.7 (Uri uses the := operator)
@dataclasses.dataclass(frozen=True)
class Uri:
    """Helper class for working with URIs."""

    scheme: str

    authority: str

    path: str

    query: str

    fragment: str

    def __post_init__(self):
        """Basic validation."""
        if self.scheme is None:
            raise ValueError("URIs must have a scheme")

        if not SCHEME.match(self.scheme):
            raise ValueError("Invalid scheme")

        if self.authority and self.path and (not self.path.startswith("/")):
            raise ValueError("Paths with an authority must start with a slash '/'")

        if self.path and self.path.startswith("//") and (not self.authority):
            raise ValueError(
                "Paths without an authority cannot start with two slashes '//'"
            )

    def __eq__(self, other):
        if type(other) is not type(self):
            return False

        if self.scheme != other.scheme:
            return False

        if self.authority != other.authority:
            return False

        if self.query != other.query:
            return False

        if self.fragment != other.fragment:
            return False

        if IS_WIN and self.scheme == "file":
            # Filepaths on windows are case in-sensitive
            if self.path.lower() != other.path.lower():
                return False

        elif self.path != other.path:
            return False

        return True

    def __hash__(self):
        if IS_WIN and self.scheme == "file":
            # Filepaths on windows are case in-sensitive
            path = self.path.lower()
        else:
            path = self.path

        return hash((self.scheme, self.authority, path, self.query, self.fragment))

    def __fspath__(self):
        """Return the file system representation of this uri.

        This makes Uri instances compatible with any function that expects an
        ``os.PathLike`` object!
        """
        # TODO: Should we raise an exception if scheme != "file"?
        return self.as_fs_path(preserve_case=True)

    def __str__(self):
        return self.as_string()

    def __truediv__(self, other):
        return self.join(other)

    @classmethod
    def create(
        cls,
        *,
        scheme: str = "",
        authority: str = "",
        path: str = "",
        query: str = "",
        fragment: str = "",
    ) -> Uri:
        """Create a uri with the given attributes."""

        if scheme in {"http", "https", "file"}:
            if not path.startswith("/"):
                path = f"/{path}"

        return cls(
            scheme=scheme,
            authority=authority,
            path=path,
            query=query,
            fragment=fragment,
        )

    @classmethod
    def parse(cls, uri: str) -> Uri:
        """Parse the given uri from its string representation."""
        scheme, authority, path, _, query, fragment = parse.urlparse(uri)
        return cls.create(
            scheme=parse.unquote(scheme),
            authority=parse.unquote(authority),
            path=parse.unquote(path),
            query=parse.unquote(query),
            fragment=parse.unquote(fragment),
        )

    def resolve(self) -> Uri:
        """Return the fully resolved version of this Uri."""

        # This operation only makes sense for file uris
        if self.scheme != "file":
            return Uri.parse(str(self))

        return Uri.for_file(pathlib.Path(self).resolve())

    @classmethod
    def for_file(cls, filepath: str | os.PathLike[str]) -> Uri:
        """Create a uri based on the given filepath."""

        fpath = os.fspath(filepath)
        if IS_WIN:
            fpath = fpath.replace("\\", "/")

        if fpath.startswith("//"):
            authority, *path = fpath[2:].split("/")
            fpath = "/".join(path)
        else:
            authority = ""

        return cls.create(scheme="file", authority=authority, path=fpath)

    @property
    def fs_path(self) -> str | None:
        """Return the equivalent fs path."""
        return self.as_fs_path()

    def where(self, **kwargs) -> Uri:
        """Return an transformed version of this uri where certain components of the uri
        have been replace with the given arguments.

        Passing a value of ``None`` will remove the given component entirely.
        """
        keys = {"scheme", "authority", "path", "query", "fragment"}
        valid_keys = keys.copy() & kwargs.keys()

        current = {k: getattr(self, k) for k in keys}
        replacements = {k: kwargs[k] for k in valid_keys}

        return Uri.create(**{**current, **replacements})

    def join(self, path: str) -> Uri:
        """Join this Uri's path component with the given path and return the resulting
        uri.

        Parameters
        ----------
        path
           The path segment to join

        Returns
        -------
        Uri
           The resulting uri
        """

        if not self.path:
            raise ValueError("This uri has no path")

        if IS_WIN:
            fs_path = self.fs_path
            if fs_path is None:
                raise ValueError("Unable to join paths, fs_path is None")

            joined = os.path.normpath(os.path.join(fs_path, path))
            new_path = self.for_file(joined).path
        else:
            new_path = os.path.normpath(os.path.join(self.path, path))

        return self.where(path=new_path)

    def as_fs_path(self, preserve_case: bool = False) -> str | None:
        """Return the file system path correspondin with this uri."""
        if self.path:
            path = _normalize_path(self.path, preserve_case)

            if self.authority and len(path) > 1:
                path = f"//{self.authority}{path}"

            # Remove the leading `/` from windows paths
            elif RE_DRIVE_LETTER_PATH.match(path):
                path = path[1:]

            if IS_WIN:
                path = path.replace("/", "\\")

            return path

        return None

    def as_string(self, encode=True) -> str:
        """Return a string representation of this Uri.

        Parameters
        ----------
        encode
           If ``True`` (the default), encode any special characters.

        Returns
        -------
        str
           The string representation of the Uri
        """

        # See: https://github.com/python/mypy/issues/10740
        encoder: Callable[[str], str] = parse.quote if encode else _replace_chars  # type: ignore[assignment]

        if authority := self.authority:
            usercred, *auth = authority.split("@")
            if len(auth) > 0:
                *user, cred = usercred.split(":")
                if len(user) > 0:
                    usercred = encoder(":".join(user)) + f":{encoder(cred)}"
                else:
                    usercred = encoder(usercred)
                authority = "@".join(auth)
            else:
                usercred = ""

            authority = authority.lower()
            *auth, port = authority.split(":")
            if len(auth) > 0:
                authority = encoder(":".join(auth)) + f":{port}"
            else:
                authority = encoder(authority)

            if usercred:
                authority = f"{usercred}@{authority}"

        scheme_separator = ""
        if authority or self.scheme == "file":
            scheme_separator = "//"

        if path := self.path:
            path = encoder(_normalize_path(path))

        if query := self.query:
            query = encoder(query)

        if fragment := self.fragment:
            fragment = encoder(fragment)

        parts = [
            f"{self.scheme}:",
            scheme_separator,
            authority if authority else "",
            path if path else "",
            f"?{query}" if query else "",
            f"#{fragment}" if fragment else "",
        ]
        return "".join(parts)


def _replace_chars(segment: str) -> str:
    """Replace a certain subset of characters in a uri segment"""
    return segment.replace("#", "%23").replace("?", "%3F")


def _normalize_path(path: str, preserve_case: bool = False) -> str:
    """Normalise the path segment of a Uri.

    Parameters
    ----------
    path
       The path to normalise.

    preserve_case
       If ``True``, preserve the case of the drive label on Windows.
       If ``False``, the drive label will be lowercased.

    Returns
    -------
    str
       The normalised path.
    """

    # normalize to fwd-slashes on windows,
    # on other systems bwd-slashes are valid
    # filename character, eg /f\oo/ba\r.txt
    if IS_WIN:
        path = path.replace("\\", "/")

    # Normalize drive paths to lower case
    if (not preserve_case) and (match := RE_DRIVE_LETTER_PATH.match(path)):
        path = match.group(1) + match.group(2).lower() + path[match.end() :]

    return path



import dataclasses
import json
import logging
import pathlib
import sys
from typing import Any

from sphinx.locale import _TranslationProxy

logger = logging.getLogger("esbonio.sphinx_agent")


def _serialize_message(obj):
    # As far as I know there's not a public function to detect just instances of a
    # dataclass...
    if dataclasses.is_dataclass(obj) and not isinstance(obj, type):
        return dataclasses.asdict(obj)

    if isinstance(obj, (_TranslationProxy, pathlib.Path)):
        return str(obj)

    if isinstance(obj, set):
        return list(obj)

    return obj


def as_json(data: Any) -> str:
    return json.dumps(data, default=_serialize_message)


def format_message(data: Any) -> str:
    content = as_json(data)
    content_length = len(content)

    return f"Content-Length: {content_length}\r\n\r\n{content}"


def send_error(id: str | int, code: int, message: str, data=None):
    send_message(
        dict(
            id=id,
            jsonrpc="2.0",
            error=dict(code=code, message=message, data=data),
        )
    )


def send_message(data: Any):
    content = format_message(data).encode("utf8")
    sys.stdout.buffer.write(content)
    sys.stdout.flush()



import asyncio
import pathlib
import sys

import pytest_lsp
from lsprotocol import types
from pytest_lsp import ClientServerConfig
from pytest_lsp import LanguageClient

SERVER_CMD = ["-m", "esbonio"]
TEST_DIR = pathlib.Path(__file__).parent.parent


@pytest_lsp.fixture(
    scope="session",
    config=ClientServerConfig(
        server_command=[sys.executable, *SERVER_CMD],
    ),
)
async def client(lsp_client: LanguageClient, uri_for, tmp_path_factory):
    """The "main" client to use for our tests."""
    workspace_uri = uri_for("workspaces", "demo")
    test_uri = workspace_uri / "index.rst"

    # Override the project configuration so that builds are written into a temporary folder
    build_dir = tmp_path_factory.mktemp("build")
    lsp_client.set_configuration({"logging": {"level": "debug"}}, section="esbonio")
    lsp_client.set_configuration(
        {
            "buildCommand": [
                "sphinx-build",
                "-M",
                "html",
                workspace_uri.fs_path,
                str(build_dir),
            ],
            "pythonCommand": [sys.executable],
        },
        section="esbonio.sphinx",
        scope_uri=str(workspace_uri),
    )

    @lsp_client.feature(types.WORKSPACE_DIAGNOSTIC_REFRESH)
    def _(lc: LanguageClient, params):
        print(f"{types.WORKSPACE_DIAGNOSTIC_REFRESH} requsted", file=sys.stderr)

    await lsp_client.initialize_session(
        types.InitializeParams(
            capabilities=types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities(
                        completion_item=types.ClientCompletionItemOptions(
                            documentation_format=[
                                types.MarkupKind.Markdown,
                                types.MarkupKind.PlainText,
                            ]
                        ),
                    ),
                    # Signal documetLink tooltip support
                    document_link=types.DocumentLinkClientCapabilities(
                        tooltip_support=True,
                    ),
                    # Signal pull diagnostic support
                    diagnostic=types.DiagnosticClientCapabilities(
                        dynamic_registration=False
                    ),
                ),
                workspace=types.WorkspaceClientCapabilities(
                    configuration=True,
                    diagnostics=types.DiagnosticWorkspaceClientCapabilities(
                        refresh_support=True,
                    ),
                ),
                # Signal workDoneProgress/create support.
                window=types.WindowClientCapabilities(
                    work_done_progress=True,
                ),
            ),
            workspace_folders=[
                types.WorkspaceFolder(uri=str(workspace_uri), name="demo"),
            ],
        )
    )

    # Open a text document to trigger sphinx client creation.
    lsp_client.text_document_did_open(
        types.DidOpenTextDocumentParams(
            text_document=types.TextDocumentItem(
                uri=str(test_uri),
                language_id="restructuredtext",
                version=0,
                text=pathlib.Path(test_uri).read_text(),
            )
        )
    )

    await lsp_client.wait_for_notification("sphinx/appCreated")

    # Save the document to trigger a build
    lsp_client.text_document_did_save(
        types.DidSaveTextDocumentParams(
            text_document=types.TextDocumentIdentifier(uri=str(test_uri))
        )
    )

    build_finished = False
    while not build_finished:
        await lsp_client.wait_for_notification("$/progress")
        report = list(lsp_client.progress_reports.values())[0][-1]
        build_finished = report.kind == "end"

    yield

    # Teardown
    await asyncio.wait_for(lsp_client.shutdown_session(), timeout=2.0)



"""This module implements the Sphinx agent.

It wraps a Sphinx application object, allowing the main language server process to
interact with it.

Whereas the language server originally had to be installed within the same Python
environment as Sphinx, the agent allows the server to run in a completely separate
Python environment and still gather the information it needs.

This is possible by taking advantage of the ``PYTHONPATH`` environment variable, using
it to expose *just this module* to the Python environment hosting Sphinx. To prevent
a potential clash of dependencies, this module is written with only what is available
in the stdlib and Sphinx itself.

Unfortunately, this does mean re-inventing some wheels, but hopefully what we gain in
portability makes it worth the trade off.
"""



from __future__ import annotations

import re
from urllib.parse import parse_qs

import pytest
from lsprotocol import types
from pytest_lsp import LanguageClient

from esbonio.server import Uri


@pytest.mark.asyncio(loop_scope="session")
async def test_preview(client: LanguageClient, uri_for):
    """Ensure that we can get ask the server to preview a given file for us."""

    workspace_uri = uri_for("workspaces", "demo")
    test_uri = workspace_uri / "rst" / "roles.rst"

    response = await client.workspace_execute_command_async(
        types.ExecuteCommandParams(
            command="esbonio.server.previewFile", arguments=[{"uri": str(test_uri)}]
        )
    )

    preview_uri = Uri.parse(response["uri"])

    assert preview_uri.scheme == "http"
    assert re.match(r"localhost:\d+", preview_uri.authority) is not None
    assert preview_uri.path == "/rst/roles.html"

    query_params = parse_qs(preview_uri.query)
    assert re.match(r"ws://localhost:\d+", query_params["ws"][0]) is not None



from __future__ import annotations

import inspect
import json
import pathlib
import re
import typing
from functools import partial
from typing import Generic
from typing import TypeVar

import attrs
from lsprotocol import types
from pygls.capabilities import get_capability

from . import Uri

T = TypeVar("T", bound=attrs.AttrsInstance)

if typing.TYPE_CHECKING:
    from collections.abc import Awaitable
    from collections.abc import Callable
    from typing import Any

    from .server import EsbonioLanguageServer

    ConfigurationCallback = Callable[["ConfigChangeEvent"], Awaitable[None] | None]


try:
    import tomllib as toml
except ImportError:
    import tomli as toml  # type: ignore[no-redef]


@attrs.define(frozen=True)
class Subscription(Generic[T]):
    """Represents a configuration subscription"""

    section: str
    """The configuration section."""

    spec: type[T]
    """The subscription's class definition."""

    context: ConfigurationContext
    """The context for this subscription."""


@attrs.define
class ConfigChangeEvent(Generic[T]):
    """Is sent to subscribers when a configuration change occurs."""

    scope: str
    """The scope at which this configuration change occured."""

    value: T
    """The latest configuration value."""

    previous: T | None = None
    """The previous configuration value, (if any)."""


VARIABLE = re.compile(r"\$\{(\w+)\}")


@attrs.define(frozen=True)
class ConfigurationContext:
    """The context in which configuration variables are evaluated."""

    file_scope: str
    """The uri of the file scope of this context."""

    workspace_scope: str
    """The uri of the workspace scope of this context."""

    @property
    def scope(self) -> str:
        """The effective scope of the config context."""
        return max([self.file_scope, self.workspace_scope], key=len)

    @property
    def scope_path(self) -> str | None:
        """The scope uri as a path."""
        uri = Uri.parse(self.scope)
        return uri.path

    @property
    def scope_fs_path(self) -> str | None:
        """The scope uri as an fs path."""
        uri = Uri.parse(self.scope)
        return uri.fs_path

    def expand(self, config: T) -> T:
        """Expand any configuration variables in the given config value."""
        for name in attrs.fields_dict(type(config)):
            value = getattr(config, name)

            # Recurse into sub-fields
            if attrs.has(type(value)):
                setattr(config, name, self.expand(value))
                continue

            # For now, we only support variables that are a string.
            if not isinstance(value, str):
                continue

            if (match := VARIABLE.match(value)) is None:
                continue

            setattr(config, name, self.expand_value(match.group(1)))

        return config

    def expand_value(self, variable: str) -> Any:
        """Return the value for the given variable name."""

        if variable == "scope":
            return self.scope

        if variable == "scopePath":
            return self.scope_path

        if variable == "scopeFsPath":
            return self.scope_fs_path

        raise ValueError(f"Undefined variable: {variable!r}")


class Configuration:
    """Manages the configuration values for the server.

    This will looks configuration in the following locations, in descending order of
    priority.

    - The server's ``initializationOptions``
    - ``workspace/configuration`` requests
    - A relevant configuration file

    **Note:** This assumes that a user's client is *either* going to be using
    ``initializationOptions`` *or* ``workspace/configuration`` not both. Any wierdness
    caused by combining both will (probably) not be fixed.

    **Note:** ``initializationOptions`` will implictly work at the global scope. There
    will be not multi-root/multi-project support using ``initializationOptions``. Use
    ``workspace/configuration`` or config files instead.
    """

    def __init__(self, server: EsbonioLanguageServer):
        self.server = server
        """The language server"""

        self.logger = server.logger.getChild("Configuration")
        """The logger instance to use"""

        self._initialization_options: dict[str, Any] = {}
        """The received initializaion options (if any)"""

        self._workspace_config: dict[str, dict[str, Any]] = {}
        """The cached workspace configuration."""

        self._file_config: dict[str, dict[str, Any]] = {}
        """The cached configuration coming from configuration files."""

        self._subscriptions: dict[
            Subscription, tuple[Any, list[ConfigurationCallback]]
        ] = {}
        """Subscriptions and their last known value"""

    @property
    def initialization_options(self):
        return self._initialization_options

    @initialization_options.setter
    def initialization_options(self, value):
        """Ensure the init options are namespaced."""
        if value is None or value == {}:
            return

        if "esbonio" not in value:
            value = dict(esbonio=value)

        self._initialization_options = value

    @property
    def converter(self):
        return self.server.converter

    @property
    def workspace(self):
        return self.server.workspace

    @property
    def supports_workspace_config(self):
        """Indicates if the client supports ``workspace/configuration`` requests."""
        return get_capability(
            self.server.client_capabilities, "workspace.configuration", False
        )

    def subscribe(
        self,
        section: str,
        spec: type[T],
        callback: ConfigurationCallback,
        scope: Uri | None = None,
    ):
        """Subscribe to updates to the given configuration section.

        Parameters
        ----------
        section
           The configuration section to subscribe to

        spec
           A class representing the configuration values of interest

        callback
           The function to call when changes are detected.

        scope
           An optional uri, specifying the scope at which to lookup the configuration.
        """
        file_scope = self._uri_to_file_scope(scope)
        workspace_scope = self._uri_to_workspace_scope(scope)

        context = ConfigurationContext(
            file_scope=file_scope, workspace_scope=workspace_scope
        )
        subscription = Subscription(section, spec, context)

        if new_subscription := subscription not in self._subscriptions:
            self.logger.debug("Creating new subscription: %s", subscription)
            self._subscriptions[subscription] = (None, [])

        value, callbacks = self._subscriptions[subscription]
        callbacks.append(callback)

        if self.server.ready.done():
            # If a value exists, notify the new callback immediately
            if value is not None:
                change_event = ConfigChangeEvent(
                    scope=subscription.context.scope,
                    value=value,
                    previous=None,
                )
                self._invoke_callback(callback, change_event)

            else:
                # Otherwise compute a value and notify all of them
                self._notify_subscription(subscription)

        elif new_subscription:
            self.server.ready.add_done_callback(
                partial(self._notify_subscription, subscription)
            )

    def _notify_subscriptions(self):
        """Notify all subscriptions about configuration changes, if necessary."""

        for subscription in self._subscriptions:
            self._notify_subscription(subscription)

    def _notify_subscription(self, subscription: Subscription, *args):
        """Check a single subscription's value and notify it, if it has changed.

        Parameters
        ----------
        subscription
           The subscription to check
        """

        previous_value, callbacks = self._subscriptions[subscription]
        value = self._get_config(
            subscription.section,
            subscription.spec,
            subscription.context,
        )

        # No need to notify if nothing has changed
        self.logger.debug("Previous: %s", previous_value)
        self.logger.debug("Current: %s", value)
        self._subscriptions[subscription] = value, callbacks

        if previous_value == value:
            return

        change_event = ConfigChangeEvent(
            scope=subscription.context.scope,
            value=value,
            previous=previous_value,
        )
        self.logger.info("%s", change_event)

        for callback in callbacks:
            self._invoke_callback(callback, change_event)

    def _invoke_callback(
        self, callback: ConfigurationCallback, change_event: ConfigChangeEvent
    ):
        """Invoke the given callback

        Parameters
        ----------
        callback
           The callback function to invoke

        change_event
           The change event to send
        """
        try:
            ret = callback(change_event)
            if inspect.iscoroutine(ret):
                self.server.run_task(ret)

        except Exception:
            self.logger.error(
                "Error in configuration callback: %s", callback, exc_info=True
            )

    def get(self, section: str, spec: type[T], scope: Uri | None = None) -> T:
        """Get the requested configuration section.

        Parameters
        ----------
        section
           The configuration section to retrieve

        spec
           A class representing the configuration values of interest

        scope
           An optional uri, specifying the scope in which to lookup the configuration.

        Returns
        -------
        T
           The requested configuration section parsed as an instance of ``T``.
        """
        context = ConfigurationContext(
            file_scope=self._uri_to_file_scope(scope),
            workspace_scope=self._uri_to_workspace_scope(scope),
        )

        return self._get_config(section, spec, context)

    def scope_for(self, uri: Uri) -> str:
        """Return the configuration scope that corresponds to the given uri.

        Parameters
        ----------
        uri
           The uri to return the scope for

        Returns
        -------
        str
           The scope corresponding with the given uri
        """
        context = ConfigurationContext(
            file_scope=self._uri_to_file_scope(uri),
            workspace_scope=self._uri_to_workspace_scope(uri),
        )

        return context.scope

    def _get_config(
        self,
        section: str,
        spec: type[T],
        context: ConfigurationContext,
    ) -> T:
        """Get the requested configuration section."""

        self.logger.debug("%s", context)

        # To keep things simple, this method assumes that all available config is already
        # cached locally. Populating the cache is handled elsewhere.
        file_config = self._file_config.get(context.file_scope, {})
        workspace_config = self._workspace_config.get(context.workspace_scope, {})

        # Combine and resolve all the config sources - order matters!
        config = _merge_configs(
            self._initialization_options, file_config, workspace_config
        )
        # self.logger.debug("Full config: %s", json.dumps(config, indent=2))

        # Extract the requested section.
        config_section = config
        for name in section.split("."):
            config_section = config_section.get(name, {})

        self.logger.debug("%s: %s", section, json.dumps(config_section, indent=2))

        try:
            value = self.converter.structure(config_section, spec)
            value = context.expand(value)
            self.logger.debug("%s", value)

            return value
        except Exception:
            self.logger.error(
                "Unable to parse configuation as '%s', using defaults",
                spec.__name__,
                exc_info=True,
            )
            return spec()

    def _uri_to_file_scope(self, uri: Uri | None) -> str:
        folder_uris = list(self._file_config.keys())
        return _uri_to_scope(folder_uris, uri)

    def _uri_to_workspace_scope(self, uri: Uri | None) -> str:
        folder_uris = [f.uri for f in self.workspace.folders.values()]

        if (root_uri := self.workspace.root_uri) is not None:
            folder_uris.append(root_uri)

        return _uri_to_scope(folder_uris, uri)

    def _discover_config_files(self) -> list[pathlib.Path]:
        """Scan the workspace for available configuration files."""
        folder_uris = {f.uri for f in self.workspace.folders.values()}

        if (root_uri := self.workspace.root_uri) is not None:
            folder_uris.add(root_uri)

        paths = []
        for uri in folder_uris:
            if (folder_path := Uri.parse(uri).fs_path) is None:
                continue

            self.logger.debug("Looking for pyproject.toml files in: '%s'", folder_path)
            for p in pathlib.Path(folder_path).glob("**/pyproject.toml"):
                self.logger.debug("Found '%s'", p)
                paths.append(p)

        return paths

    def update_file_configuration(self, paths: list[pathlib.Path] | None = None):
        """Update the internal cache of configuration coming from files.

        Parameters
        ----------
        paths
           A list of filepaths to read from.
           If not set, this method will scan each workspace folder for relevant files.
        """
        if paths is None:
            paths = self._discover_config_files()

        for path in paths:
            try:
                data = toml.loads(path.read_text())
                config = dict(esbonio=data.get("tool", {}).get("esbonio", {}))
                scope = str(Uri.for_file(path.parent))

                self._file_config[scope] = config
                self.logger.debug(
                    "File '%s' configuration: %s", scope, json.dumps(config, indent=2)
                )
            except Exception:
                self.logger.error(
                    "Unable to read configuration file: '%s'", exc_info=True
                )

        self._notify_subscriptions()

    async def update_workspace_configuration(self):
        """Update the internal cache of the client's workspace configuration."""
        if not self.supports_workspace_config:
            return

        # Request configuration at the global scope, and at each workspace.
        scopes = [None] + [f.uri for f in self.workspace.folders.values()]
        if (root_uri := self.workspace.root_uri) is not None and root_uri not in scopes:
            scopes.append(root_uri)

        params = types.ConfigurationParams(
            items=[
                types.ConfigurationItem(section="esbonio", scope_uri=scope)
                for scope in scopes
            ]
        )
        self.logger.debug(
            "workspace/configuration: %s",
            json.dumps(self.converter.unstructure(params), indent=2),
        )

        try:
            results = await self.server.workspace_configuration_async(params)
        except Exception:
            self.logger.error("Unable to get workspace configuration", exc_info=True)
            return

        for scope, result in zip(scopes, results, strict=False):
            self.logger.debug(
                "Workspace '%s' configuration: %s", scope, json.dumps(result, indent=2)
            )
            # result can be `None`
            result = result or {}

            if "esbonio" not in result:
                result = dict(esbonio=result)

            self._workspace_config[scope or ""] = result

        self._notify_subscriptions()


def _uri_to_scope(known_scopes: list[str], uri: Uri | None) -> str:
    """Convert the given uri to a scope or the empty string if none could be found.

    Parameters
    ----------
    known_scopes
       The collection of known scopes.

    uri
       The uri to convert

    Returns
    -------
    str
        The scope to use
    """
    if uri is None:
        return ""

    uri = uri.resolve()
    known_uris = [str(Uri.parse(u).resolve()) for u in known_scopes]

    candidates = [scope for scope in known_uris if str(uri).startswith(scope)]
    if len(candidates) == 0:
        return ""

    # Return the most specific match.
    return sorted(candidates, key=len, reverse=True)[0]


def _merge_configs(*configs: dict[str, Any]):
    """Recursively merge all the given configuration sources together.

    The last config given takes precedence.
    """
    final = {}
    all_keys: set[str] = set()

    for c in configs:
        all_keys.update(c.keys())

    for k in all_keys:
        values = [
            v for c in configs if (v := c.get(k, attrs.NOTHING)) is not attrs.NOTHING
        ]

        # Do we need to recurse?
        if isinstance(values[-1], dict):
            # Be sure to only pass dictionary values to _merge_configs.
            final[k] = _merge_configs(*[v for v in values if isinstance(v, dict)])
        else:
            final[k] = values[-1]

    return final



from __future__ import annotations

import hashlib
import logging
import os
import pathlib
import sys
import typing
from unittest import mock

import pytest
from sphinx import version_info as sphinx_version

from esbonio.sphinx_agent.config import SphinxConfig
from esbonio.sphinx_agent.log import DiagnosticFilter
from esbonio.sphinx_agent.log import source_to_uri_and_linum
from esbonio.sphinx_agent.types import Uri

if typing.TYPE_CHECKING:
    from typing import Any


logger = logging.getLogger(__name__)


sphinx_lt_81 = sphinx_version[0] < 8 or (
    sphinx_version[0] == 8 and sphinx_version[1] < 1
)


def application_args(**kwargs) -> dict[str, Any]:
    defaults = {
        "confoverrides": {},
        "freshenv": False,
        "keep_going": False,
        "parallel": 1,
        "tags": [],
        "verbosity": 0,
        "warningiserror": False,
    }

    for arg in ("srcdir", "outdir", "confdir", "doctreedir"):
        if arg in kwargs and (value := kwargs[arg]) is not None:
            kwargs[arg] = str(pathlib.Path(value).resolve())

    # Order matters, kwargs will override any keys found in defaults.
    return {**defaults, **kwargs}


@pytest.mark.parametrize(
    "args, expected",
    [
        (
            ["-M", "html", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-M", "html", "src", "out", "-C"],
            application_args(
                srcdir="src",
                confdir=None,
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-M", "latex", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "latex"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="latex",
            ),
        ),
        (
            ["-M", "html", "src", "out", "-E"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                freshenv=True,
            ),
        ),
        (
            ["-M", "html", "src", "out", "-c", "conf"],
            application_args(
                srcdir="src",
                confdir="conf",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-M", "html", "src", "out", "-d", "doctreedir"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir="doctreedir",
                buildername="html",
            ),
        ),
        (
            ["-M", "html", "src", "out", "-Dkey=value", "-Danother=v"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                confoverrides=dict(
                    key="value",
                    another="v",
                ),
            ),
        ),
        (
            ["-M", "html", "src", "out", "-Akey=value", "-Aanother=v"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                confoverrides={
                    "html_context.key": "value",
                    "html_context.another": "v",
                },
            ),
        ),
        (
            ["-M", "html", "src", "out", "-j", "4"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                parallel=4,
            ),
        ),
        (
            ["-M", "html", "src", "out", "-n"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                confoverrides=dict(nitpicky=True),
            ),
        ),
        (
            # quiet = True is handled by the logging framework
            ["-M", "html", "src", "out", "-q"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
            ),
        ),
        (
            # silent = True is handled by the logging framework
            ["-M", "html", "src", "out", "-Q"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-M", "html", "src", "out", "-t", "tag1"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                tags=["tag1"],
            ),
        ),
        (
            ["-M", "html", "src", "out", "-t", "tag1", "-t", "tag2"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                tags=["tag1", "tag2"],
            ),
        ),
        (
            ["-M", "html", "src", "out", "-v"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                verbosity=1,
            ),
        ),
        (
            ["-M", "html", "src", "out", "-vvv"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                verbosity=3,
            ),
        ),
        (
            ["-M", "html", "src", "out", "-W"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                warningiserror=True,
            ),
        ),
        (
            ["-M", "html", "src", "out", "-W", "--keep-going"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.path.join("out", "html"),
                doctreedir=os.path.join("out", "doctrees"),
                buildername="html",
                warningiserror=True,
                # --keep-going ignored since v8.1
                # https://github.com/sphinx-doc/sphinx/pull/12743/files#diff-4aea2ac365ab5325b530ac42efc67feac98db110e7f943ea13b5cf88f4260e59
                keep_going=sphinx_lt_81,
            ),
        ),
        (
            ["-b", "html", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-b", "html", "src", "out", "-C"],
            application_args(
                srcdir="src",
                confdir=None,
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-b", "latex", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="latex",
            ),
        ),
        (
            ["-b", "html", "-E", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                freshenv=True,
            ),
        ),
        (
            ["-b", "html", "-c", "conf", "src", "out"],
            application_args(
                srcdir="src",
                confdir="conf",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-b", "html", "-Dkey=value", "-Danother=v", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                confoverrides=dict(
                    key="value",
                    another="v",
                ),
            ),
        ),
        (
            ["-b", "html", "-Akey=value", "-Aanother=v", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                confoverrides={
                    "html_context.key": "value",
                    "html_context.another": "v",
                },
            ),
        ),
        (
            ["-b", "html", "-d", "doctreedir", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir="doctreedir",
                buildername="html",
            ),
        ),
        (
            ["-b", "html", "-j", "4", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                parallel=4,
            ),
        ),
        (
            ["-b", "html", "-n", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                confoverrides={
                    "nitpicky": True,
                },
            ),
        ),
        (
            ["-b", "html", "-q", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-b", "html", "-Q", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
            ),
        ),
        (
            ["-b", "html", "-t", "tag1", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                tags=["tag1"],
            ),
        ),
        (
            ["-b", "html", "-t", "tag1", "-t", "tag2", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                tags=["tag1", "tag2"],
            ),
        ),
        (
            ["-b", "html", "-v", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                verbosity=1,
            ),
        ),
        (
            ["-b", "html", "-vvv", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                verbosity=3,
            ),
        ),
        (
            ["-b", "html", "-W", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                warningiserror=True,
            ),
        ),
        (
            ["-b", "html", "-W", "--keep-going", "src", "out"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir="out",
                doctreedir=os.path.join("out", ".doctrees"),
                buildername="html",
                warningiserror=True,
                # --keep-going ignored since v8.1
                # https://github.com/sphinx-doc/sphinx/pull/12743/files#diff-4aea2ac365ab5325b530ac42efc67feac98db110e7f943ea13b5cf88f4260e59
                keep_going=sphinx_lt_81,
            ),
        ),
    ],
)
def test_cli_arg_handling(args: list[str], expected: dict[str, Any]):
    """Ensure that we can convert ``sphinx-build`` to the correct Sphinx application
    options."""
    config = SphinxConfig.fromcli(args)
    assert config is not None

    actual = config.to_application_args({})

    # pytest overrides stderr on windows, so if we were to put `sys.stderr` in the
    # `expected` dict this test would fail as `sys.stderr` inside a test function has a
    # different value.
    #
    # So, let's test for it here instead
    assert actual.pop("status") == sys.stderr
    assert actual.pop("warning") == sys.stderr

    assert expected == actual


@pytest.mark.parametrize(
    "args, expected, context",
    [
        (
            ["-M", "html", "src", "${defaultBuildDir}"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.sep + os.path.join("path", "to", "cache", "<HASH>", "html"),
                doctreedir=(
                    os.sep + os.path.join("path", "to", "cache", "<HASH>", "doctrees")
                ),
                buildername="html",
            ),
            {
                "cacheDir": os.sep + os.sep.join(["path", "to", "cache"]),
            },
        ),
        (
            ["-M", "html", "src", "${defaultBuildDir}", "-C"],
            application_args(
                srcdir="src",
                confdir=None,
                outdir=os.sep + os.path.join("path", "to", "cache", "<HASH>", "html"),
                doctreedir=(
                    os.sep + os.path.join("path", "to", "cache", "<HASH>", "doctrees")
                ),
                buildername="html",
            ),
            {
                "cacheDir": os.sep + os.sep.join(["path", "to", "cache"]),
            },
        ),
        (
            ["-M", "html", "src", "${defaultBuildDir}"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.sep + os.path.join("path", "to", ".cache", "<HASH>", "html"),
                doctreedir=(
                    os.sep + os.path.join("path", "to", ".cache", "<HASH>", "doctrees")
                ),
                buildername="html",
            ),
            {
                "cacheDir": os.sep + os.sep.join(["path", "to", ".cache"]),
            },
        ),
        (
            ["-b", "html", "src", "${defaultBuildDir}"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.sep + os.sep.join(["path", "to", "cache", "<HASH>"]),
                doctreedir=(
                    os.sep + os.path.join("path", "to", "cache", r"<HASH>", ".doctrees")
                ),
                buildername="html",
            ),
            {
                "cacheDir": os.sep + os.sep.join(["path", "to", "cache"]),
            },
        ),
        (
            ["-b", "html", "src", "${defaultBuildDir}", "-C"],
            application_args(
                srcdir="src",
                confdir=None,
                outdir=os.sep + os.sep.join(["path", "to", "cache", "<HASH>"]),
                doctreedir=(
                    os.sep + os.path.join("path", "to", "cache", r"<HASH>", ".doctrees")
                ),
                buildername="html",
            ),
            {
                "cacheDir": os.sep + os.sep.join(["path", "to", "cache"]),
            },
        ),
        (
            ["-b", "html", "src", "${defaultBuildDir}"],
            application_args(
                srcdir="src",
                confdir="src",
                outdir=os.sep + os.sep.join(["path", "to", ".cache", "<HASH>"]),
                doctreedir=(
                    os.sep
                    + os.path.join("path", "to", ".cache", r"<HASH>", ".doctrees")
                ),
                buildername="html",
            ),
            {
                "cacheDir": os.sep + os.sep.join(["path", "to", ".cache"]),
            },
        ),
    ],
)
def test_cli_default_build_dir(
    args: list[str], expected: dict[str, Any], context: dict[str, str]
):
    """Ensure that we can handle ``${defaultBuildDir}`` variable correctly."""
    config = SphinxConfig.fromcli(args)
    assert config is not None

    actual = config.to_application_args(context)

    # pytest overrides stderr on windows, so if we were to put `sys.stderr` in the
    # `expected` dict this test would fail as `sys.stderr` inside a test function has a
    # different value.
    #
    # So, let's test for it here instead
    assert actual.pop("status") == sys.stderr
    assert actual.pop("warning") == sys.stderr

    # Compute the expected hash now that the confdir has been resolved
    dirname = config.conf_dir or config.src_dir
    ehash = hashlib.md5(dirname.encode()).hexdigest()
    expected["outdir"] = expected["outdir"].replace("<HASH>", ehash)
    expected["doctreedir"] = expected["doctreedir"].replace("<HASH>", ehash)

    assert expected == actual


ROOT = pathlib.Path(__file__).parent.parent / "sphinx-extensions" / "workspace"
PY_PATH = ROOT / "code" / "diagnostics.py"
CONF_PATH = ROOT / "sphinx-extensions" / "conf.py"
RST_PATH = ROOT / "sphinx-extensions" / "index.rst"
INC_PATH = ROOT / "sphinx-extensions" / "_include_me.txt"
REL_INC_PATH = os.path.relpath(INC_PATH)


@pytest.mark.parametrize(
    "location, expected",
    [
        (f"{RST_PATH}", (Uri.for_file(RST_PATH), None)),
        (f"{RST_PATH}:", (Uri.for_file(RST_PATH), None)),
        (f"{RST_PATH}:3", (Uri.for_file(RST_PATH), 3)),
        (f"{REL_INC_PATH}:12", (Uri.for_file(INC_PATH), 12)),
        (
            f"{PY_PATH}:docstring of esbonio.sphinx_agent.log.DiagnosticFilter:3",
            (Uri.for_file(PY_PATH), 22),
        ),
        (f"internal padding after {RST_PATH}:34", (Uri.for_file(RST_PATH), 34)),
        (f"internal padding before {RST_PATH}:34", (Uri.for_file(RST_PATH), 34)),
    ],
)
def test_source_to_uri_linum(location: str, expected: tuple[str, int | None]):
    """Ensure we can correctly determine a dianostic's location based on the string we
    get from sphinx."""

    mockpath = f"{DiagnosticFilter.__module__}.inspect.getsourcelines"
    with mock.patch(mockpath, return_value=([""], 20)):
        actual = source_to_uri_and_linum(location)

    assert actual == expected



from __future__ import annotations

import re
import typing
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from .lsp import Location

if typing.TYPE_CHECKING:
    from collections.abc import Callable
    from typing import TypeVar

    T = TypeVar("T")
    JsonLoader = Callable[[str, type[T]], T]


MYST_DIRECTIVE: re.Pattern = re.compile(
    r"""
    (\s*)                             # directives can be indented
    (?P<directive>
      ```(`*)?                        # directives start with at least 3 ` chars
      [{]?                            # followed by an opening brace
      (?P<name>[^}]+)?                # directives have a name
      [}]?                            # directives are closed with a closing brace
    )
    (\s+(?P<argument>.*?)\s*$)?       # directives may take an argument
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse partial and complete MyST directives.

This does **not** include any options or content that may be included with the
initial declaration.
"""


RST_DIRECTIVE: re.Pattern = re.compile(
    r"""
    (\s*)                             # directives can be indented
    (?P<directive>
      \.\.                            # directives start with a comment
      [ ]?                            # followed by a space
      (?P<substitution>\|             # this could be a substitution definition
        (?P<substitution_text>[^|]+)?
      \|?)?
      [ ]?
      (?P<name>([\w-]|:(?!:))+)?      # directives have a name
      (::)?                           # directives end with '::'
    )
    ([\s]+(?P<argument>.*?)\s*$)?     # directives may take an argument
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse partial and complete directives.

This does **not** include any options or content that may be included underneath
the initial declaration. A number of named capture groups are available.

``name``
   The name of the directive, not including the domain prefix.

``directive``
   Everything that makes up a directive, from the initial ``..`` up to and including the
   ``::`` characters.

``argument``
   All argument text.

``substitution``
   If the directive is part of a substitution definition, this group will contain
"""


RST_DIRECTIVE_OPTION: re.Pattern = re.compile(
    r"""
    (?P<indent>\s+)       # directive options must be indented
    (?P<option>
      :                   # options start with a ':'
      (?P<name>[\w-]+)?   # options have a name
      :?                  # options end with a ':'
    )
    (\s*
      (?P<value>.*)       # options can have a value
    )?
    """,
    re.VERBOSE,
)
"""A regular expression used to detect and parse partial and complete directive options.

A number of named capture groups are available

``name``
   The name of the option

``option``
   The name of the option including the surrounding ``:`` characters.

``indent``
   The whitespace characters making preceeding the initial ``:`` character

``value``
   The value passed to the option

"""


@dataclass
class Directive:
    """Represents a directive."""

    @dataclass
    class ArgumentProvider:
        """An argument provider instance."""

        name: str
        """The name of the provider."""

        kwargs: dict[str, Any] = field(default_factory=dict)
        """Arguments to pass to the argument provider"""

    name: str
    """The name of the directive, as the user would type in an rst file."""

    implementation: str | None
    """The dotted name of the directive's implementation."""

    location: Location | None = field(default=None)
    """The location of the directive's implementation, if known"""

    argument_providers: list[ArgumentProvider] = field(default_factory=list)
    """The list of argument providers that can be used with this directive."""

    def to_db(
        self, dumps: Callable[[Any], str]
    ) -> tuple[str, str | None, str | None, str | None]:
        """Convert this directive to its database representation"""

        providers = None
        if len(self.argument_providers) > 0:
            providers = dumps(self.argument_providers)

        location = dumps(self.location) if self.location is not None else None
        return (self.name, self.implementation, location, providers)

    @classmethod
    def from_db(
        cls,
        load_as: JsonLoader,
        name: str,
        implementation: str | None,
        location: str | None,
        providers: str | None,
    ) -> Directive:
        """Create a directive from its database representation"""

        loc = load_as(location, Location) if location is not None else None
        argument_providers = (
            load_as(providers, list[Directive.ArgumentProvider])
            if providers is not None
            else []
        )

        return cls(
            name=name,
            implementation=implementation,
            location=loc,
            argument_providers=argument_providers,
        )



from __future__ import annotations

import asyncio
import logging
import typing
from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler

from esbonio import server
from esbonio.server import Uri

if typing.TYPE_CHECKING:
    from typing import Any

    from .config import PreviewConfig


class RequestHandler(SimpleHTTPRequestHandler):
    def __init__(self, *args, logger: logging.Logger, directory: str, **kwargs) -> None:
        self.logger = logger
        super().__init__(*args, directory=directory, **kwargs)

    def translate_path(self, path: str) -> str:
        result = super().translate_path(path)
        # self.logger.debug("Translate: '%s' -> '%s'", path, result)
        return result

    def end_headers(self):
        # Instruct the client not to cache the response.
        #
        # See: https://github.com/swyddfa/esbonio/issues/951
        #      https://github.com/swyddfa/esbonio/issues/987
        self.send_header("Cache-Control", "no-store")
        return super().end_headers()

    def log_message(self, format: str, *args: Any) -> None:
        self.logger.debug(format, *args)


class RequestHandlerFactory:
    """Class for dynamically producing request handlers.

    ``HTTPServer`` works by taking a "request handler" class and creating an instance of
    it for every request it receives. By making this class callable, we can dynamically
    produce a request handler based on the current situation.
    """

    def __init__(self, logger: logging.Logger, build_uri: Uri | None = None):
        self.logger = logger
        self.build_uri = build_uri

    def __call__(self, *args, **kwargs):
        if self.build_uri is None:
            raise ValueError("No build directory set")

        if (build_dir := self.build_uri.fs_path) is None:
            raise ValueError(
                "Unable to determine build dir from uri: '%s'", self.build_uri
            )

        return RequestHandler(*args, logger=self.logger, directory=build_dir, **kwargs)


class PreviewServer:
    """The http server that serves the built content."""

    def __init__(self, logger: logging.Logger, config: PreviewConfig, executor: Any):
        self.config = config
        """The current configuration."""

        self.logger = logger.getChild("PreviewServer")
        """The logger instance to use."""

        self._handler_factory = RequestHandlerFactory(self.logger)
        """Factory for producing http request handlers."""

        self._startup_task: asyncio.Task | None = None
        """Task that resolves once the server is ready."""

        self._executor: Any = executor
        """The executor in which to run the http server."""

        self._future: asyncio.Future | None = None
        """The future representing the http server's "task"."""

        self._server: HTTPServer | None = None
        """The http server itself."""

    def __await__(self):
        """Makes the server await-able"""
        if self._startup_task is None:
            self._startup_task = asyncio.create_task(self.start())

        return self._startup_task.__await__()

    @property
    def port(self):
        if self._server is None:
            return 0

        return self._server.server_port

    @property
    def build_uri(self):
        return self._handler_factory.build_uri

    @build_uri.setter
    def build_uri(self, value):
        self._handler_factory.build_uri = value

    async def start(self):
        """Start the server."""

        # Yes, this method does not need to be async. However, making it async means it
        # aligns well with the pattern we've established in other components.

        self._server = HTTPServer(
            (self.config.bind, self.config.http_port), self._handler_factory
        )

        loop = asyncio.get_running_loop()
        self._future = loop.run_in_executor(self._executor, self._server.serve_forever)

        return self

    def stop(self):
        """Stop the server."""
        if self._server is not None:
            self.logger.debug("Shutting down preview HTTP server")
            self._server.shutdown()

        if self._future is not None:
            self.logger.debug("Cancelling HTTP future: %s", self._future.cancel())


def make_http_server(
    esbonio: server.EsbonioLanguageServer, config: PreviewConfig
) -> PreviewServer:
    return PreviewServer(esbonio.logger, config, esbonio.thread_pool)



from __future__ import annotations

import asyncio
import collections
import inspect
import logging
import platform
import traceback
import typing
from concurrent.futures import Future
from typing import TypeVar
from uuid import uuid4

import cattrs
from lsprotocol import types
from pygls.capabilities import get_capability
from pygls.lsp.server import LanguageServer
from pygls.workspace import TextDocument
from pygls.workspace import Workspace

from . import Uri
from ._configuration import Configuration

if typing.TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Coroutine
    from typing import Any

    from .feature import LanguageFeature

__version__ = "1.0.0"
T = TypeVar("T")
LF = TypeVar("LF", bound="LanguageFeature")


class EsbonioWorkspace(Workspace):
    """A modified version of pygls' workspace that ensures uris are always resolved."""

    def get_text_document(self, doc_uri: str) -> TextDocument:
        uri = str(Uri.parse(doc_uri).resolve())
        return super().get_text_document(uri)

    def put_text_document(
        self, text_document: types.TextDocumentItem, notebook_uri: str | None = None
    ):
        text_document.uri = str(Uri.parse(text_document.uri).resolve())
        return super().put_text_document(text_document)

    def remove_text_document(self, doc_uri: str):
        doc_uri = str(Uri.parse(doc_uri).resolve())
        return super().remove_text_document(doc_uri)

    def update_text_document(
        self,
        text_doc: types.VersionedTextDocumentIdentifier,
        change: types.TextDocumentContentChangeEvent,
    ):
        text_doc.uri = str(Uri.parse(text_doc.uri).resolve())
        return super().update_text_document(text_doc, change)


class EsbonioLanguageServer(LanguageServer):
    """The Esbonio language server"""

    def __init__(self, logger: logging.Logger | None = None, *args, **kwargs):
        if "name" not in kwargs:
            kwargs["name"] = "esbonio"

        if "version" not in kwargs:
            kwargs["version"] = __version__

        super().__init__(*args, **kwargs)

        self._diagnostics: dict[tuple[str, Uri], list[types.Diagnostic]] = {}
        """Where we store and manage diagnostics."""

        self._loaded_extensions: dict[str, Any] = {}
        """Record of server modules that have been loaded."""

        self._features: dict[type[LanguageFeature], LanguageFeature] = {}
        """The collection of language features registered with the server."""

        self._ready: Future[bool] = Future()
        """Indicates if the server is ready."""

        self._tasks: set[asyncio.Task] = set()
        """Used to hold running tasks"""

        self.logger = logger or logging.getLogger(__name__)
        """The logger instance to use."""

        self.configuration = Configuration(self)
        """Manages the fetching of configuration values."""

    def __iter__(self):
        return iter(self._features.items())

    @property
    def ready(self) -> Future[bool]:
        return self._ready

    @property
    def converter(self) -> cattrs.Converter:
        """The cattrs converter instance we should use."""
        return self.protocol._converter

    def _finish_task(self, task: asyncio.Task[Any]):
        """Cleanup a finished task."""
        self.logger.debug("Task finished: %s", task)
        self._tasks.discard(task)

        if (exc := task.exception()) is not None:
            self.logger.error(
                "Error in async task\n%s",
                traceback.format_exception(type(exc), exc, exc.__traceback__),
            )

    def run_task(self, coro: Coroutine, *, name: str | None = None) -> asyncio.Task:
        """Convert a given coroutine into a task and ensure it is executed."""

        task = asyncio.create_task(coro, name=name)
        self.logger.debug("Scheduled task: %s", task)
        task.add_done_callback(self._finish_task)

        self._tasks.add(task)
        return task

    def initialize(self, params: types.InitializeParams):
        self.logger.info(
            "Initialising esbonio v%s, using Python v%s on %s",
            __version__,
            platform.python_version(),
            platform.platform(aliased=True, terse=True),
        )
        if (client := params.client_info) is not None:
            self.logger.info("Language client: %s %s", client.name, client.version)

        # TODO: Propose patch to pygls for providing custom Workspace implementations.
        self.protocol._workspace = EsbonioWorkspace(
            self.workspace.root_uri,
            self.workspace._sync_kind,
            list(self.workspace.folders.values()),
        )

        self.configuration.initialization_options = params.initialization_options

    async def initialized(self, params: types.InitializedParams):
        self.configuration.update_file_configuration()

        await asyncio.gather(
            self.configuration.update_workspace_configuration(),
            self._register_did_change_configuration_handler(),
            self._register_did_change_watched_files_handler(),
        )

    def lsp_shutdown(self, params: None):
        """Called when the server is instructed to ``shutdown`` by the client."""

    def load_extension(self, name: str, setup: Callable):
        """Load the given setup function as an extension.

        If an extension with the given ``name`` already exists, the given setup function
        will be ignored.

        The ``setup`` function can declare dependencies in the form of type
        annotations.

        .. code-block:: python

           from esbonio.lsp.roles import Roles
           from esbonio.lsp.sphinx import SphinxLanguageServer


           def esbonio_setup(rst: SphinxLanguageServer, roles: Roles): ...

        In this example the setup function is requesting instances of the
        :class:`~esbonio.lsp.sphinx.SphinxLanguageServer` and the
        :class:`~esbonio.lsp.roles.Roles` language feature.

        Parameters
        ----------
        name
           The name to give the extension

        setup
           The setup function to call
        """

        if name in self._loaded_extensions:
            self.logger.debug("Skipping extension '%s', already loaded", name)
            return

        arguments = _get_setup_arguments(self, setup, name)
        if not arguments:
            return

        try:
            setup(**arguments)

            self.logger.debug("Loaded extension '%s'", name)
            self._loaded_extensions[name] = setup
        except Exception:
            self.logger.error("Unable to load extension '%s'", name, exc_info=True)

    def add_feature(self, feature: LanguageFeature):
        """Register a language feature with the server.

        Parameters
        ----------
        feature
           The language feature
        """
        feature_cls = type(feature)
        if feature_cls in self._features:
            name = f"{feature_cls.__module__}.{feature_cls.__name__}"
            raise RuntimeError(f"Feature '{name}' has already been registered")

        self._features[feature_cls] = feature

    def get_feature(self, feature_cls: type[LF]) -> LF | None:
        """Returns the requested language feature if it exists, otherwise it returns
        ``None``.

        Parameters
        ----------
        feature_cls
           The class definiion of the feature to retrieve
        """
        return self._features.get(feature_cls, None)  # type: ignore

    def clear_diagnostics(self, source: str, uri: Uri | None = None) -> None:
        """Clear diagnostics from the given source.

        Parameters
        ----------
        source:
           The source from which to clear diagnostics.
        uri:
           If given, clear diagnostics from within just this uri. Otherwise, all
           diagnostics from the given source are cleared.
        """

        for key in self._diagnostics.keys():
            clear_source = source == key[0]
            clear_uri = uri == key[1] or uri is None

            if clear_source and clear_uri:
                self._diagnostics[key] = []

    def add_diagnostics(self, source: str, uri: Uri, diagnostic: types.Diagnostic):
        """Add a diagnostic to the given source and uri.

        Parameters
        ----------
        source
           The source the diagnostics are from
        uri
           The uri the diagnostics are associated with
        diagnostic
           The diagnostic to add
        """
        key = (source, uri)
        self._diagnostics.setdefault(key, []).append(diagnostic)

    def set_diagnostics(
        self, source: str, uri: Uri, diagnostics: list[types.Diagnostic]
    ) -> None:
        """Set the diagnostics for the given source and uri.

        Parameters
        ----------
        source:
           The source the diagnostics are from
        uri:
           The uri the diagnostics are associated with
        diagnostics:
           The diagnostics themselves
        """
        self._diagnostics[(source, uri)] = diagnostics

    def get_language_at(self, document: TextDocument, position: types.Position) -> str:
        """Return the language at the given location"""
        language = document.language_id

        if language in {"rst", "restructuredtext"}:
            return "rst"

        if language in {"markdown"}:
            return "markdown"

        return ""

    def sync_diagnostics(self) -> None:
        """Update the client with the currently stored diagnostics using
        a ``textDocument/publishDiagnostics`` notification.

        When using the pull diagnostics model, this triggers a
        ``workspace/diagnostic/refresh`` request instead.
        """
        diagnostic_provider = self.server_capabilities.diagnostic_provider
        workspace_diagnostic = get_capability(
            self.client_capabilities,
            "workspace.diagnostics",
            types.DiagnosticWorkspaceClientCapabilities(),
        )

        if diagnostic_provider is not None:
            if workspace_diagnostic.refresh_support:
                # Signal to the client that is should ask again for more diagnostics
                self.workspace_diagnostic_refresh(None)
            return

        uris = {uri for _, uri in self._diagnostics.keys()}
        diagnostics = {uri: DiagnosticList() for uri in uris}

        for (_, uri), diags in self._diagnostics.items():
            diagnostics[uri].extend(diags)

        for uri, diag_list in diagnostics.items():
            self.logger.debug("Publishing %d diagnostics for: %s", len(diag_list), uri)
            self.text_document_publish_diagnostics(
                types.PublishDiagnosticsParams(uri=str(uri), diagnostics=diag_list.data)
            )

    async def _register_did_change_watched_files_handler(self):
        """Register the server's handler for ``workspace/didChangeWatchedFiles``."""
        capabilities = self.client_capabilities
        registration_supported = get_capability(
            capabilities,
            "workspace.did_change_watched_files.dynamic_registration",
            False,
        )
        if not registration_supported:
            self.logger.info(
                "Client does not support dynamic registration of '%s' handlers, "
                "server might not be able to react to configuration changes.",
                types.WORKSPACE_DID_CHANGE_WATCHED_FILES,
            )
            return

        try:
            await self.client_register_capability_async(
                types.RegistrationParams(
                    registrations=[
                        types.Registration(
                            id=str(uuid4()),
                            method=types.WORKSPACE_DID_CHANGE_WATCHED_FILES,
                            register_options=types.DidChangeWatchedFilesRegistrationOptions(
                                watchers=[
                                    types.FileSystemWatcher(
                                        glob_pattern="**/pyproject.toml"
                                    )
                                ]
                            ),
                        ),
                    ]
                )
            )
            self.logger.debug(
                "Registered '%s' handler", types.WORKSPACE_DID_CHANGE_WATCHED_FILES
            )
        except Exception:
            self.logger.error(
                "Unable to register '%s' handler",
                types.WORKSPACE_DID_CHANGE_WATCHED_FILES,
                exc_info=True,
            )

    async def _register_did_change_configuration_handler(self):
        """Register the server's handler for ``workspace/didChangeConfiguration``.

        The spec says that in order to receive these notifications we need to
        dynamically register our capability to process them. (Though I think some
        editors send them regardless.)
        """
        capabilities = self.client_capabilities
        registration_supported = get_capability(
            capabilities,
            "workspace.did_change_configuration.dynamic_registration",
            False,
        )
        if not registration_supported:
            self.logger.info(
                "Client does not support dynamic registration of '%s' handlers, "
                "server might not be able to react to configuration changes.",
                types.WORKSPACE_DID_CHANGE_CONFIGURATION,
            )
            return

        try:
            await self.client_register_capability_async(
                types.RegistrationParams(
                    registrations=[
                        types.Registration(
                            id=str(uuid4()),
                            method=types.WORKSPACE_DID_CHANGE_CONFIGURATION,
                        ),
                    ]
                )
            )
            self.logger.debug(
                "Registered '%s' handler", types.WORKSPACE_DID_CHANGE_CONFIGURATION
            )
        except Exception:
            self.logger.error(
                "Unable to register '%s' handler",
                types.WORKSPACE_DID_CHANGE_CONFIGURATION,
                exc_info=True,
            )


class DiagnosticList(collections.UserList):
    """A list type dedicated to holding diagnostics.

    This is mainly to ensure that only one instance of a diagnostic ever gets
    reported.
    """

    def append(self, item: types.Diagnostic):
        if not isinstance(item, types.Diagnostic):
            raise TypeError("Expected Diagnostic")

        for existing in self.data:
            fields = [
                existing.range == item.range,
                existing.message == item.message,
                existing.severity == item.severity,
                existing.code == item.code,
                existing.source == item.source,
            ]

            if all(fields):
                # Item already added, nothing to do.
                return

        self.data.append(item)


def _get_setup_arguments(
    server: EsbonioLanguageServer, setup: Callable, modname: str
) -> dict[str, Any] | None:
    """Given a setup function, try to construct the collection of arguments to pass to
    it.
    """
    annotations = typing.get_type_hints(setup)
    parameters = {
        p.name: annotations[p.name]
        for p in inspect.signature(setup).parameters.values()
    }

    args = {}
    for name, type_ in parameters.items():
        if issubclass(server.__class__, type_):
            args[name] = server
            continue

        from .feature import LanguageFeature

        if issubclass(type_, LanguageFeature):
            # Try and obtain an instance of the requested language feature.
            feature = server.get_feature(type_)
            if feature is not None:
                args[name] = feature
                continue

            server.logger.debug(
                "Skipping extension '%s', server missing requested feature: '%s'",
                modname,
                type_,
            )
            return None

        server.logger.error(
            "Skipping extension '%s', parameter '%s' has unsupported type: '%s'",
            modname,
            name,
            type_,
        )
        return None

    return args



from __future__ import annotations

import asyncio
import dataclasses
import json
import logging
import re
import sys
import threading
import traceback
import typing
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import TypeVar

# This has to be called here before any imports from sphinx are made.
# Otherwise we'll be too late and miss the chance to patch things properly
from .patches import patch_sphinx

patch_sphinx()

from .handlers import SphinxHandler  # type: ignore
from .util import send_error

HANDLERS = SphinxHandler()
logger = logging.getLogger(__name__)


T = TypeVar("T")


def parse_message(obj: dict, cls: type[T]) -> T:
    """Convert a raw dict into the given type"""

    if dataclasses.is_dataclass(cls):
        kwargs = {}
        fields = typing.get_type_hints(cls)

        for key, value in obj.items():
            kwargs[key] = parse_message(value, fields[key])

        return cls(**kwargs)  # type: ignore[return-value]

    return obj  # type: ignore[return-value]


def handle_message(data: bytes):
    message = json.loads(data.decode("utf8"))

    method = message.get("method", None)
    if not method:
        # TODO: Return a valid JSON-RPC error
        raise TypeError("Invalid message")

    result = HANDLERS.get(method)
    if result is None:
        # TODO: Return a valid JSON-RPC error
        raise TypeError(f"Unknown method: '{method}'")

    type_, handler = result
    obj: Any = parse_message(message, type_)
    try:
        handler(obj)
    except Exception as e:
        msg_id = message.get("id", None)
        if msg_id is not None:
            send_error(
                id=msg_id,
                code=-32603,
                message=f"{e}",
                data=dict(traceback=traceback.format_exc()),
            )


async def main_loop(loop, executor, stop_event, rfile, proxy):
    """Originally taken from ``pygls``"""

    CONTENT_LENGTH_PATTERN = re.compile(rb"^Content-Length: (\d+)\r\n$")

    # Initialize message buffer
    content_length = 0

    while not stop_event.is_set() and not rfile.closed:
        # Read a header line
        header = await loop.run_in_executor(executor, rfile.readline)
        if not header:
            break

        # Extract content length if possible
        if not content_length:
            match = CONTENT_LENGTH_PATTERN.fullmatch(header)
            if match:
                content_length = int(match.group(1))
                logger.debug("Content length: %s", content_length)

        # Check if all headers have been read (as indicated by an empty line \r\n)
        if content_length and not header.strip():
            # Read body
            body = await loop.run_in_executor(executor, rfile.read, content_length)
            if not body:
                break

            proxy(body)

            # Reset the buffer
            content_length = 0


async def main():
    loop = asyncio.get_running_loop()
    event = threading.Event()
    executor = ThreadPoolExecutor(max_workers=2)

    await main_loop(loop, executor, event, sys.stdin.buffer, handle_message)
